# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module contains the main types of the ``structure`` subpackage:
:class:`Atom`, :class:`AtomArray` and :class:`AtomArrayStack`. 
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann"
__all__ = ["Atom", "AtomArray", "AtomArrayStack",
           "array", "stack", "repeat", "from_template", "coord"]

import numbers
import abc
import numpy as np
from .bonds import BondList
from ..copyable import Copyable


class _AtomArrayBase(Copyable, metaclass=abc.ABCMeta):
    """
    Private base class for :class:`AtomArray` and
    :class:`AtomArrayStack`.
    It implements functionality for annotation arrays and also
    rudimentarily for coordinates.
    """
    
    def __init__(self, length):
        """
        Create the annotation arrays
        """
        self._annot = {}
        self._array_length = length
        self._coord = None
        self._bonds = None
        self._box = None
        self.add_annotation("chain_id", dtype="U4")
        self.add_annotation("res_id", dtype=int)
        self.add_annotation("ins_code", dtype="U1")
        self.add_annotation("res_name", dtype="U3")
        self.add_annotation("hetero", dtype=bool)
        self.add_annotation("atom_name", dtype="U6")
        self.add_annotation("element", dtype="U2")
        
    def array_length(self):
        """
        Get the length of the atom array.
        
        This value is equivalent to the length of each annotation array.
        For :class:`AtomArray` it is the same as ``len(array)``.
        
        Returns
        -------
        length : int
            Length of the array(s).
        """
        return self._array_length

    @property
    @abc.abstractmethod
    def shape(self):
        """
        Tuple of array dimensions.

        This property contains the current shape of the object.

        Returns
        -------
        shape : tuple of int
            Shape of the object.
        """
        return 
        
    def add_annotation(self, category, dtype):
        """
        Add an annotation category, if not already existing.
        
        Initially the new annotation is filled with the *zero*
        representation of the given type.
        
        Parameters
        ----------
        category : str
            The annotation category to be added.
        dtype : type or str
            A type instance or a valid *NumPy* *dtype* string.
            Defines the type of the annotation
        
        See Also
        --------
        set_annotation
        """
        if category not in self._annot:
            self._annot[str(category)] = np.zeros(self._array_length,
                                                  dtype=dtype)
            
    def del_annotation(self, category):
        """
        Removes an annotation category.
        
        Parameters
        ----------
        category : str
            The annotation category to be removed.
        """
        if category in self._annot:
            del self._annot[str(category)]
            
    def get_annotation(self, category):
        """
        Return an annotation array.
        
        Parameters
        ----------
        category : str
            The annotation category to be returned.
            
        Returns
        -------
        array : ndarray
            The annotation array.
        """
        if category not in self._annot:
            raise ValueError(
                f"Annotation category '{category}' is not existing"
            )
        return self._annot[category]
    
    def set_annotation(self, category, array):
        """
        Set an annotation array. If the annotation category does not
        exist yet, the category is created.
        
        Parameters
        ----------
        category : str
            The annotation category to be set.
        array : ndarray or None
            The new value of the annotation category. The size of the
            array must be the same as the array length.
        """
        if len(array) != self._array_length:
            raise IndexError(
                f"Expected array length {self._array_length}, "
                f"but got {len(array)}"
            )
        if category in self._annot:
            # Keep the dtype if the annotation already exists
            self._annot[category] = np.asarray(
                array, dtype=self._annot[category].dtype
            )
        else:
            self._annot[category] = np.asarray(array)
        
    def get_annotation_categories(self):
        """
        Return a list containing all annotation array categories.
            
        Returns
        -------
        categories : list
            The list containing the names of each annotation array.
        """
        return list(self._annot.keys())
            
    def _subarray(self, index):
        # Index is one dimensional (boolean mask, index array)
        new_coord = self._coord[..., index, :]
        new_length = new_coord.shape[-2]
        if isinstance(self, AtomArray):
            new_object = AtomArray(new_length)
        elif isinstance(self, AtomArrayStack):
            new_depth = new_coord.shape[-3]
            new_object = AtomArrayStack(new_depth, new_length)
        new_object._coord = new_coord
        if self._bonds is not None:
            new_object._bonds = self._bonds[index]
        if self._box is not None:
            new_object._box = self._box
        for annotation in self._annot:
            new_object._annot[annotation] = (self._annot[annotation]
                                             .__getitem__(index))
        return new_object
        
    def _set_element(self, index, atom):
        try:
            if isinstance(index, (numbers.Integral, np.ndarray)):
                for name in self._annot:
                    self._annot[name][index] = atom._annot[name]
                self._coord[..., index, :] = atom.coord
            else:
                raise TypeError(
                    f"Index must be integer, not '{type(index).__name__}'"
                )
        except KeyError:
            raise KeyError("The annotations of the 'Atom' are incompatible")
        
    def _del_element(self, index):
        if isinstance(index, numbers.Integral):
            for name in self._annot:
                self._annot[name] = np.delete(self._annot[name], index, axis=0)
            self._coord = np.delete(self._coord, index, axis=-2)
            self._array_length = self._coord.shape[-2]
            if self._bonds is not None:
                mask = np.ones(self._bonds.get_atom_count(), dtype=bool)
                mask[index] = False
                self._bonds = self._bonds[mask]
        else:
            raise TypeError(
                    f"Index must be integer, not '{type(index).__name__}'"
                )
    
    def equal_annotations(self, item):
        """
        Check, if this object shares equal annotation arrays with the
        given :class:`AtomArray` or :class:`AtomArrayStack`.
        
        Parameters
        ----------
        item : AtomArray or AtomArrayStack
            The object to compare the annotation arrays with.
        
        Returns
        -------
        equality : bool
            True, if the annotation arrays are equal.
        """
        if not isinstance(item, _AtomArrayBase):
            return False
        if not self.equal_annotation_categories(item):
            return False
        for name in self._annot:
            if not np.array_equal(self._annot[name], item._annot[name]):
                return False
        return True
    
    def equal_annotation_categories(self, item):
        """
        Check, if this object shares equal annotation array categories
        with the given :class:`AtomArray` or :class:`AtomArrayStack`.
        
        Parameters
        ----------
        item : AtomArray or AtomArrayStack
            The object to compare the annotation arrays with.
        
        Returns
        -------
        equality : bool
            True, if the annotation array names are equal.
        """
        return sorted(self._annot.keys()) == sorted(item._annot.keys())
    
    def __getattr__(self, attr):
        """
        If the attribute is an annotation, the annotation is returned
        from the dictionary.
        Exposes coordinates.
        """
        if attr == "coord":
            return self._coord
        if attr == "bonds":
            return self._bonds
        if attr == "box":
            return self._box
        # Call method of 'object' superclass to avoid infinite recursive
        # calls of '__getattr__()'
        elif attr in super().__getattribute__("_annot"):
            return self._annot[attr]
        else:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{attr}'"
            )
        
    def __setattr__(self, attr, value):
        """
        If the attribute is an annotation, the :attr:`value` is saved
        to the annotation in the dictionary.
        Exposes coordinates.
        :attr:`value` must have same length as :func:`array_length()`.
        """
        if attr == "coord":
            if not isinstance(value, np.ndarray):
                raise TypeError("Value must be ndarray of floats")
            if isinstance(self, AtomArray):
                if value.ndim != 2:
                    raise ValueError(
                        "A 2-dimensional ndarray is expected "
                        "for an AtomArray"
                )
            elif isinstance(self, AtomArrayStack):
                if value.ndim != 3:
                    raise ValueError(
                        "A 3-dimensional ndarray is expected "
                        "for an AtomArrayStack"
                )
            if value.shape[-2] != self._array_length:
                raise ValueError(
                    f"Expected array length {self._array_length}, "
                    f"but got {len(value)}"
                )
            if value.shape[-1] != 3:
                raise TypeError("Expected 3 coordinates for each atom")
            super().__setattr__("_coord", value.astype(np.float32, copy=False))
        
        elif attr == "bonds":
            if isinstance(value, BondList):
                if value.get_atom_count() != self._array_length:
                    raise ValueError(
                        f"Array length is {self._array_length}, "
                        f"but bond list has {value.get_atom_count()} atoms"
                    )
                super().__setattr__("_bonds", value)
            elif value is None:
                # Remove bond list
                super().__setattr__("_bonds", None)
            else:
                raise TypeError("Value must be 'BondList'")
        
        elif attr == "box":
            if isinstance(value, np.ndarray):
                if isinstance(self, AtomArray):
                    if value.ndim != 2:
                        raise ValueError(
                            "A 2-dimensional ndarray is expected "
                            "for an AtomArray"
                    )
                else:   # AtomArrayStack
                    if value.ndim != 3:
                        raise ValueError(
                            "A 3-dimensional ndarray is expected "
                            "for an AtomArrayStack"
                    )
                if value.shape[-2:] != (3,3):
                    raise TypeError("Box must be a 3x3 matrix (three vectors)")
                box = value.astype(np.float32, copy=False)
                super().__setattr__("_box", box)
            elif value is None:
                # Remove box
                super().__setattr__("_box", None)
            else:
                raise TypeError("Box must be ndarray of floats or None")
        
        elif attr == "_annot":
            super().__setattr__(attr, value)
        elif attr in self._annot:
            self.set_annotation(attr, value)
        else:
            super().__setattr__(attr, value)
            
    def __dir__(self):
        attr = super().__dir__()
        attr.append("coord")
        attr.append("bonds")
        attr.append("box")
        for name in self._annot.keys():
            attr.append(name)
        return attr
    
    def __eq__(self, item):
        """
        See Also
        --------
        equal_annotations
        """
        if not self.equal_annotations(item):
            return False
        if self._bonds != item._bonds:
            return False
        if self._box is None:
            if item._box is not None:
                return False
        else:
            if not np.array_equal(self._box, item._box):
                return False
        return np.array_equal(self._coord, item._coord)
    
    def __len__(self):
        """
        The length of the annotation arrays.
        
        Returns
        -------
        length : int
            Length of the annotation arrays.
        """
        return self._array_length
    
    def __add__(self, array):
        if type(self) != type(array):
            raise TypeError("Can only concatenate two arrays or two stacks")
        # Create either new array or stack, depending of the own type
        if isinstance(self, AtomArray):
            concat = AtomArray(length = self._array_length+array._array_length)
        if isinstance(self, AtomArrayStack):
            concat = AtomArrayStack(self.stack_depth(),
                                    self._array_length + array._array_length)
        
        concat._coord = np.concatenate((self._coord, array.coord), axis=-2)
        
        # Transfer only annotations,
        # which are existent in both operands
        arr_categories = list(array._annot.keys())
        for category in self._annot.keys():
            if category in arr_categories:
                annot = self._annot[category]
                arr_annot = array._annot[category]
                concat._annot[category] = np.concatenate((annot,arr_annot))
        
        # Concatenate bonds lists,
        # if at least one of them contains bond information
        if self._bonds is not None or array._bonds is not None:
            bonds1 = self._bonds
            bonds2 = array._bonds
            if bonds1 is None:
               bonds1 = BondList(self._array_length)
            if bonds2 is None:
                bonds2 = BondList(array._array_length)
            concat._bonds = bonds1 + bonds2
        
        # Copy box
        if self._box is not None:
            concat._box = np.copy(self._box)
        return concat
    
    def __copy_fill__(self, clone):
        super().__copy_fill__(clone)
        self._copy_annotations(clone)
        clone._coord = np.copy(self._coord)
    
    def _copy_annotations(self, clone):
        for name in self._annot:
            clone._annot[name] = np.copy(self._annot[name])
        if self._box is not None:
            clone._box = np.copy(self._box)
        if self._bonds is not None:
            clone._bonds = self._bonds.copy()
    

class Atom(Copyable):
    """
    A representation of a single atom.
    
    The coordinates an annotations can be accessed directly.
    A detailed description of each annotation category can be viewed
    :doc:`here </apidoc/biotite.structure>`.
    
    Parameters
    ----------
    coord: list or ndarray
        The x, y and z coordinates.
    kwargs
        Atom annotations as key value pair.
    
    Attributes
    ----------
    {annot} : scalar
        Annotations for this atom.
    coord : ndarray, dtype=float
        ndarray containing the x, y and z coordinate of the atom.
    shape : tuple of int
        Shape of the object.
        In case of an :class:`Atom`, the tuple is empty.
    
    Examples
    --------
    
    >>> atom = Atom([1,2,3], chain_id="A")
    >>> atom.atom_name = "CA"
    >>> print(atom.atom_name)
    CA
    >>> print(atom.coord)
    [1. 2. 3.]
        
    """
    
    def __init__(self, coord, **kwargs):
        self._annot = {}
        self._annot["chain_id"] = ""
        self._annot["res_id"] = 0
        self._annot["ins_code"] = ""
        self._annot["res_name"] = ""
        self._annot["hetero"] = False
        self._annot["atom_name"] = ""
        self._annot["element"] = ""
        if "kwargs" in kwargs:
            # kwargs are given directly as dictionary
            kwargs = kwargs["kwargs"]
        for name, annotation in kwargs.items():
            self._annot[name] = annotation
        coord = np.array(coord, dtype=np.float32)
        # Check if coord contains x,y and z coordinates
        if coord.shape != (3,):
            raise ValueError("Position must be ndarray with shape (3,)")
        self.coord = coord

    def __repr__(self):
        """Represent Atom as a string for debugging."""
        annot = 'chain_id="' + self._annot["chain_id"] + '"'
        annot = annot + ', res_id=' + str(self._annot["res_id"])
        annot = annot + ', ins_code="' + self._annot["ins_code"] + '"'
        annot = annot + ', res_name="' + self._annot["res_name"] + '"'
        annot = annot + ', hetero=' + str(self._annot["hetero"])
        annot = annot + ', atom_name="' + self._annot["atom_name"] + '"'
        annot = annot + ', element="' + self._annot["element"] + '"'
        return f'Atom(np.{np.array_repr(self.coord)}, {annot})'

    @property
    def shape(self):
        return ()
        
    def __getattr__(self, attr):
        if attr in super().__getattribute__("_annot"):
            return self._annot[attr]
        else:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{attr}'"
            )
        
    def __setattr__(self, attr, value):
        if attr == "_annot":
            super().__setattr__(attr, value)
        elif attr == "coord":
            super().__setattr__(attr, value)
        else:
            self._annot[attr] = value
    
    def __str__(self):
        hetero = "HET" if self.hetero else ""
        return f"{hetero:3} {self.chain_id:3} " \
               f"{self.res_id:5d}{self.ins_code:1} {self.res_name:3} " \
               f"{self.atom_name:6} {self.element:2}     " \
               f"{self.coord[0]:8.3f} " \
               f"{self.coord[1]:8.3f} " \
               f"{self.coord[2]:8.3f}"
    
    def __eq__(self, item):
        if not isinstance(item, Atom):
            return False
        if not np.array_equal(self.coord, item.coord):
            return False
        if self._annot.keys() != item._annot.keys():
            return False
        for name in self._annot:
            if self._annot[name] != item._annot[name]:
                return False
        return True
    
    def __ne__(self, item):
        return not self == item
    
    def __copy_create__(self):
        return Atom(self.coord, **self._annot)

    
class AtomArray(_AtomArrayBase):
    """
    An array representation of a model consisting of multiple atoms.
    
    An :class:`AtomArray` can be seen as a list of :class:`Atom`
    instances.
    Instead of using directly a list, this class uses an *NumPy*
    :class:`ndarray` for each annotation category and the coordinates.
    These
    coordinates can be accessed directly via the :attr:`coord`
    attribute.
    The annotations are accessed either via the category as attribute
    name or the :func:`get_annotation()`, :func:`set_annotation()`
    method.
    Usage of custom annotations is achieved via :func:`add_annotation()`
    or :func:`set_annotation()`.
    A detailed description of each annotation category can be viewed
    :doc:`here </apidoc/biotite.structure>`.
    
    In order to get an an subarray of an :class:`AtomArray`,
    *NumPy* style indexing is used.
    This includes slices, boolean arrays, index arrays and even
    *Ellipsis* notation.
    Using a single integer as index returns a single :class:`Atom`
    instance.
    
    Inserting or appending an :class:`AtomArray` to another
    :class:`AtomArray` is done with the '+' operator.
    Only the annotation categories, which are existing in both arrays,
    are transferred to the new array.

    Optionally, an :class:`AtomArray` can store chemical bond
    information via a :class:`BondList` object.
    It can be accessed using the :attr:`bonds` attribute.
    If no bond information is available, :attr:`bonds` is ``None``.
    Consequently the bond information can be removed from the
    :class:`AtomArray`, by setting :attr:`bonds` to ``None``.
    When indexing the :class:`AtomArray` the atom indices in the
    associated :class:`BondList` are updated as well, hence the indices
    in the :class:`BondList` will always point to the same atoms.
    If two :class:`AtomArray` instances are concatenated, the resulting
    :class:`AtomArray` will contain the merged :class:`BondList` if at
    least one of the operands contains bond information.

    The :attr:`box` attribute contains the box vectors of the unit cell
    or the MD simulation box, respectively.
    Hence, it is a *3 x 3* *ndarray* with the vectors in the last
    dimension.
    If no box is provided, the attribute is ``None``.
    Setting the :attr:`box` attribute to ``None`` means removing the
    box from the atom array.

    Parameters
    ----------
    length : int
        The fixed amount of atoms in the array.
    
    Attributes
    ----------
    {annot} : ndarray
        Multiple n-length annotation arrays.
    coord : ndarray, dtype=float, shape=(n,3)
        ndarray containing the x, y and z coordinate of the
        atoms.
    bonds : BondList or None
        A :class:`BondList`, specifying the indices of atoms
        that form a chemical bond.
    box : ndarray, dtype=float, shape=(3,3) or None
        The surrounding box. May represent a MD simulation box
        or a crystallographic unit cell.
    shape : tuple of int
        Shape of the atom array.
        The single value in the tuple is
        the length of the atom array.
    
    Examples
    --------
    Creating an atom array from atoms:
    
    >>> atom1 = Atom([1,2,3], chain_id="A")
    >>> atom2 = Atom([2,3,4], chain_id="A")
    >>> atom3 = Atom([3,4,5], chain_id="B")
    >>> atom_array = array([atom1, atom2, atom3])
    >>> print(atom_array.array_length())
    3
    
    Accessing an annotation array:
    
    >>> print(atom_array.chain_id)
    ['A' 'A' 'B']
        
    Accessing the coordinates:
    
    >>> print(atom_array.coord)
    [[1. 2. 3.]
     [2. 3. 4.]
     [3. 4. 5.]]
    
    *NumPy* style filtering:
    
    >>> atom_array = atom_array[atom_array.chain_id == "A"]
    >>> print(atom_array.array_length())
    2
        
    Inserting an atom array:
        
    >>> insert = array([Atom([7,8,9], chain_id="C")])
    >>> atom_array = atom_array[0:1] + insert + atom_array[1:2]
    >>> print(atom_array.chain_id)
    ['A' 'C' 'A']
    """
    
    def __init__(self, length):
        super().__init__(length)
        if length is None:
            self._coord = None
        else:
            self._coord = np.full((length, 3), np.nan, dtype=np.float32)

    def __repr__(self):
        """Represent AtomArray as a string for debugging."""
        atoms = ''
        for i in range(0, self.array_length()):
            if len(atoms) == 0:
                atoms = '\n\t' + self.get_atom(i).__repr__()
            else:
                atoms = atoms + ',\n\t' + self.get_atom(i).__repr__()
        return f'array([{atoms}\n])'

    @property
    def shape(self):
        """
        Tuple of array dimensions.

        This property contains the current shape of the
        :class:`AtomArray`.

        Returns
        -------
        shape : tuple of int
            Shape of the array.
            The single value in the tuple is
            the :func:`array_length()`.

        See Also
        --------
        array_length
        """
        return self.array_length(),

    def get_atom(self, index):
        """
        Obtain the atom instance of the array at the specified index.
        
        The same as ``array[index]``, if `index` is an integer.
        
        Parameters
        ----------
        index : int
            Index of the atom.
        
        Returns
        -------
        atom : Atom
            Atom at position `index`. 
        """
        kwargs = {}
        for name, annotation in self._annot.items():
            kwargs[name] = annotation[index]
        return Atom(coord = self._coord[index], kwargs=kwargs)
    
    def __iter__(self):
        """
        Iterate through the array.
        
        Yields
        ------
        atom : Atom
        """
        i = 0
        while i < len(self):
            yield self.get_atom(i)
            i += 1
    
    def __getitem__(self, index):
        """
        Obtain a subarray or the atom instance at the specified index.
        
        Parameters
        ----------
        index : object
            All index types *NumPy* accepts, are valid.
        
        Returns
        -------
        sub_array : Atom or AtomArray
            If `index` is an integer an :class:`Atom` instance is
            returned.
            Otherwise an :class:`AtomArray` with reduced length is
            returned.
        """
        if isinstance(index, numbers.Integral):
            return self.get_atom(index)
        elif isinstance(index, tuple):
            if len(index) == 2 and index[0] is Ellipsis:
                # If first index is "...", just ignore the first index
                return self.__getitem__(index[1])
            else:
                raise IndexError(
                    "'AtomArray' does not accept multidimensional indices"
                )
        else:
            return self._subarray(index)
        
    def __setitem__(self, index, atom):
        """
        Set the atom at the specified array position.
        
        Parameters
        ----------
        index : int
            The position, where the atom is set.
        atom : Atom
            The atom to be set.
        """
        self._set_element(index, atom)
        
    def __delitem__(self, index):
        """
        Deletes the atom at the specified array position.
        
        Parameters
        ----------
        index : int
            The position where the atom should be deleted.
        """
        self._del_element(index)
        
    def __len__(self):
        """
        The length of the array.
        
        Returns
        -------
        length : int
            Length of the array.
        """
        return self.array_length()
    
    def __eq__(self, item):
        """
        Check if the array equals another :class:`AtomArray`.
        
        Parameters
        ----------
        item : object
            Object to campare the array with.
        
        Returns
        -------
        equal : bool
            True, if `item` is an :class:`AtomArray`
            and all its attribute arrays equals the ones of this object.
        """
        if not super().__eq__(item):
            return False
        if not isinstance(item, AtomArray):
            return False
        return True
    
    def __str__(self):
        """
        Get a string representation of the array.
        
        Each line contains the attributes of one atom.
        """
        return "\n".join([str(atom) for atom in self])
    
    def __copy_create__(self):
        return AtomArray(self.array_length())


class AtomArrayStack(_AtomArrayBase):
    """
    A collection of multiple :class:`AtomArray` instances, where each
    atom array has equal annotation arrays.
    
    Effectively, this means that each atom is occuring in every array in
    the stack at differing coordinates. This situation arises e.g. in
    NMR-elucidated or simulated structures. Since the annotations are
    equal for each array, the annotation arrays are 1-D, while the
    coordinate array is 3-D (m x n x 3).
    A detailed description of each annotation category can be viewed
    :doc:`here </apidoc/biotite.structure>`.
    
    Indexing works similar to :class:`AtomArray`, with the difference,
    that two index dimensions are possible:
    The first index dimension specifies the array(s), the second index
    dimension specifies the atoms in each array (same as the index
    in :class:`AtomArray`).
    Using a single integer as first dimension index returns a single
    :class:`AtomArray` instance.
    
    Concatenation of atoms for each array in the stack is done using the
    '+' operator. For addition of atom arrays onto the stack use the
    :func:`stack()` method.

    The :attr:`box` attribute has the shape *m x 3 x 3*, as the cell
    might be different for each frame in the atom array stack.
    
    Parameters
    ----------
    depth : int
        The fixed amount of arrays in the stack. When indexing, this is
        the length of the first dimension.
        
    length : int
        The fixed amount of atoms in each array in the stack. When
        indexing, this is the length of the second dimension.
    
    Attributes
    ----------
    {annot} : ndarray, shape=(n,)
        Mutliple n-length annotation arrays.
    coord : ndarray, dtype=float, shape=(m,n,3)
        ndarray containing the x, y and z coordinate of the
        atoms.
    bonds: BondList or None
        A :class:`BondList`, specifying the indices of atoms
        that form a chemical bond.
    box: ndarray, dtype=float, shape=(m,3,3) or None
        The surrounding box. May represent a MD simulation box
        or a crystallographic unit cell.
    shape : tuple of int
        Shape of the stack.
        The numbers correspond to the stack depth
        and array length, respectively.
    
    See also
    --------
    AtomArray
    
    Examples
    --------
    Creating an atom array stack from two arrays:
    
    >>> atom1 = Atom([1,2,3], chain_id="A")
    >>> atom2 = Atom([2,3,4], chain_id="A")
    >>> atom3 = Atom([3,4,5], chain_id="B")
    >>> atom_array1 = array([atom1, atom2, atom3])
    >>> print(atom_array1.coord)
    [[1. 2. 3.]
     [2. 3. 4.]
     [3. 4. 5.]]
    >>> atom_array2 = atom_array1.copy()
    >>> atom_array2.coord += 3
    >>> print(atom_array2.coord)
    [[4. 5. 6.]
     [5. 6. 7.]
     [6. 7. 8.]]
    >>> array_stack = stack([atom_array1, atom_array2])
    >>> print(array_stack.coord)
    [[[1. 2. 3.]
      [2. 3. 4.]
      [3. 4. 5.]]
    <BLANKLINE>
     [[4. 5. 6.]
      [5. 6. 7.]
      [6. 7. 8.]]]
    """
    
    def __init__(self, depth, length):
        super().__init__(length)
        if depth == None or length == None:
            self._coord = None
        else:
            self._coord = np.full((depth, length, 3), np.nan, dtype=np.float32)

    def __repr__(self):
        """Represent AtomArrayStack as a string for debugging."""
        arrays = ''
        for i in range(0, self.stack_depth()):
            if len(arrays) == 0:
                arrays = '\n\t' + self.get_array(i).__repr__()
            else:
                arrays = arrays + ',\n\t' + self.get_array(i).__repr__()
        return f'stack([{arrays}\n])'

    def get_array(self, index):
        """
        Obtain the atom array instance of the stack at the specified
        index.
        
        The same as ``stack[index]``, if `index` is an integer.
        
        Parameters
        ----------
        index : int
            Index of the atom array.
        
        Returns
        -------
        array : AtomArray
            AtomArray at position `index`. 
        """
        array = AtomArray(self.array_length())
        for name in self._annot:
            array._annot[name] = self._annot[name]
        array._coord = self._coord[index]
        if self._bonds is not None:
            array._bonds = self._bonds.copy()
        if self._box is not None:
            array._box = self._box[index]

        return array
    
    def stack_depth(self):
        """
        Get the depth of the stack.
        
        This value represents the amount of atom arrays in the stack.
        It is the same as ``len(array)``.
        
        Returns
        -------
        length : int
            Length of the array(s).
        """
        return len(self)

    @property
    def shape(self):
        """
        Tuple of array dimensions.

        This property contains the current shape of the
        :class:`AtomArrayStack`.

        Returns
        -------
        shape : tuple of int
            Shape of the stack.
            The numbers correspond to the :func:`stack_depth()`
            and :func:`array_length()`, respectively.
        """
        return self.stack_depth(), self.array_length()

    def __iter__(self):
        """
        Iterate through the array.
        
        Yields
        ------
        array : AtomArray
        """
        i = 0
        while i < len(self):
            yield self.get_array(i)
            i += 1
            
    def __getitem__(self, index):
        """
        Obtain the atom array instance or an substack at the specified
        index.
        
        Parameters
        ----------
        index : object
            All index types *NumPy* accepts are valid.
        
        Returns
        -------
        sub_array : AtomArray or AtomArrayStack
            If `index` is an integer an :class:`AtomArray` instance is
            returned.
            Otherwise an :class:`AtomArrayStack` with reduced depth and
            length is returned.
            In case the index is a tuple(int, int) an :class:`Atom`
            instance is returned.  
        """
        if isinstance(index, numbers.Integral):
            return self.get_array(index)
        elif isinstance(index, tuple):
            if len(index) != 2:
                raise IndexError(
                    "'AtomArrayStack' does not accept an index "
                    "with more than two dimensions"
                )
            if isinstance(index[0], numbers.Integral):
                array = self.get_array(index[0])
                return array.__getitem__(index[1])
            else:
                if isinstance(index[1], numbers.Integral):
                    # Prevent reduction in dimensionality
                    # in second dimension
                    new_stack = self._subarray(slice(index[1], index[1]+1))
                else:
                    new_stack = self._subarray(index[1])
                if index[0] is not Ellipsis:
                    new_stack._coord = new_stack._coord[index[0]]
                    if new_stack._box is not None:
                        new_stack._box = new_stack._box[index[0]]
                return new_stack
        else:
            new_stack = AtomArrayStack(depth=0, length=self.array_length())
            self._copy_annotations(new_stack)
            new_stack._coord = self._coord[index]
            if self._box is not None:
                new_stack._box = self._box[index]
            return new_stack
            
    
    def __setitem__(self, index, array):
        """
        Set the atom array at the specified stack position.
        
        The array and the stack must have equal annotation arrays.
        
        Parameters
        ----------
        index : int
            The position, where the array atom is set.
        array : AtomArray
            The atom array to be set.
        """
        if not self.equal_annotations(array):
            raise ValueError(
                "The stack and the array have unequal annotations"
            )
        if self.bonds != array.bonds:
            raise ValueError(
                "The stack and the array have unequal bonds"
            )
        if isinstance(index, numbers.Integral):
            self.coord[index] = array.coord
            if self.box is not None:
                self.box[index] = array.box
        else:
            raise TypeError(
                f"Index must be integer, not '{type(index).__name__}'"
            )
        
    def __delitem__(self, index):
        """
        Deletes the atom array at the specified stack position.
        
        Parameters
        ----------
        index : int
            The position where the atom array should be deleted.
        """
        if isinstance(index, numbers.Integral):
            self._coord = np.delete(self._coord, index, axis=0)
        else:
            raise TypeError(
                f"Index must be integer, not '{type(index).__name__}'"
            )
    
    def __len__(self):
        """
        The depth of the stack, i.e. the amount of models.
        
        Returns
        -------
        depth : int
            depth of the array.
        """
        # length is determined by length of coord attribute
        return self._coord.shape[0]
    
    def __eq__(self, item):
        """
        Check if the array equals another :class:`AtomArray`
        
        Parameters
        ----------
        item : object
            Object to campare the array with.
        
        Returns
        -------
        equal : bool
            True, if `item` is an :class:`AtomArray`
            and all its attribute arrays equals the ones of this object.
        """
        if not super().__eq__(item):
            return False
        if not isinstance(item, AtomArrayStack):
            return False
        return True
    
    def __str__(self):
        """
        Get a string representation of the stack.
        
        :class:`AtomArray` strings eparated by blank lines
        and a line indicating the index.
        """
        string = ""
        for i, array in enumerate(self):
            string += "Model " + str(i+1) + "\n"
            string += str(array) + "\n" + "\n"
        return string
    
    def __copy_create__(self):
        return AtomArrayStack(self.stack_depth(), self.array_length())


def array(atoms):
    """
    Create an :class:`AtomArray` from a list of :class:`Atom`.
    
    Parameters
    ----------
    atoms : iterable object of Atom
        The atoms to be combined in an array.
        All atoms must share the same annotation categories.
    
    Returns
    -------
    array : AtomArray
        The listed atoms as array.
    
    Examples
    --------
    
    Creating an atom array from atoms:
    
    >>> atom1 = Atom([1,2,3], chain_id="A")
    >>> atom2 = Atom([2,3,4], chain_id="A")
    >>> atom3 = Atom([3,4,5], chain_id="B")
    >>> atom_array = array([atom1, atom2, atom3])
    >>> print(atom_array)
        A       0                       1.000    2.000    3.000
        A       0                       2.000    3.000    4.000
        B       0                       3.000    4.000    5.000
    """
    # Check if all atoms have the same annotation names
    # Equality check requires sorting
    names = sorted(atoms[0]._annot.keys())
    for i, atom in enumerate(atoms):
        if sorted(atom._annot.keys()) != names:
            raise ValueError(
                f"The atom at index {i} does not share the same "
                f"annotation categories as the atom at index 0"
            )
    array = AtomArray(len(atoms))
    # Add all (also optional) annotation categories
    for name in names:
        array.add_annotation(name, dtype=type(atoms[0]._annot[name])) 
    # Add all atoms to AtomArray
    for i in range(len(atoms)):
        for name in names:
            array._annot[name][i] = atoms[i]._annot[name]
        array._coord[i] = atoms[i].coord
    return array


def stack(arrays):
    """
    Create an :class:`AtomArrayStack` from a list of :class:`AtomArray`.
    
    Parameters
    ----------
    arrays : iterable object of AtomArray
        The atom arrays to be combined in a stack.
        All atom arrays must have an equal number of atoms and equal
        annotation arrays.
    
    Returns
    -------
    stack : AtomArrayStack
        The stacked atom arrays.
    
    Examples
    --------
    Creating an atom array stack from two arrays:
    
    >>> atom1 = Atom([1,2,3], chain_id="A")
    >>> atom2 = Atom([2,3,4], chain_id="A")
    >>> atom3 = Atom([3,4,5], chain_id="B")
    >>> atom_array1 = array([atom1, atom2, atom3])
    >>> print(atom_array1.coord)
    [[1. 2. 3.]
     [2. 3. 4.]
     [3. 4. 5.]]
    >>> atom_array2 = atom_array1.copy()
    >>> atom_array2.coord += 3
    >>> print(atom_array2.coord)
    [[4. 5. 6.]
     [5. 6. 7.]
     [6. 7. 8.]]
    >>> array_stack = stack([atom_array1, atom_array2])
    >>> print(array_stack.coord)
    [[[1. 2. 3.]
      [2. 3. 4.]
      [3. 4. 5.]]
    <BLANKLINE>
     [[4. 5. 6.]
      [5. 6. 7.]
      [6. 7. 8.]]]
    """
    array_count = 0
    ref_array = None
    for i, array in enumerate(arrays):
        if ref_array is None:
            ref_array = array
        array_count += 1
        # Check if all arrays share equal annotations
        if not array.equal_annotations(ref_array):
            raise ValueError(
                f"The annotations of the atom array at index {i} are not "
                f"equal to the annotations of the atom array at index 0"
            )
    array_stack = AtomArrayStack(array_count, ref_array.array_length())
    for name, annotation in ref_array._annot.items():
        array_stack._annot[name] = annotation
    coord_list = [array._coord for array in arrays] 
    array_stack._coord = np.stack(coord_list, axis=0)
    # Take bond list from first array
    array_stack._bonds = ref_array._bonds
    # When all atom arrays provide a box, copy the boxes
    if all([array.box is not None for array in arrays]):
        array_stack.box = np.array([array.box for array in arrays])
    return array_stack


def repeat(atoms, coord):
    """
    Repeat atoms (:class:`AtomArray` or :class:`AtomArrayStack`)
    multiple times in the same model with different coordinates.

    Parameters
    ----------
    atoms : AtomArray, shape=(n,) or AtomArrayStack, shape=(m,n)
        The atoms to be repeated.
    coord : ndarray, dtype=float, shape=(k,n,3) or shape=(k,m,n,3)
        The coordinates to be used fr the repeated atoms.
        The length of first dimension determines the number of repeats.
        If `atoms` is an :class:`AtomArray` 3 dimensions, otherwise
        4 dimensions are required.
    
    Returns
    -------
    repeated: AtomArray, shape=(n*k,) or AtomArrayStack, shape=(m,n*k)
        The repeated atoms.
        Whether an :class:`AtomArray` or an :class:`AtomArrayStack` is
        returned depends on the input `atoms`.
    
    Examples
    --------

    >>> atoms = array([
    ...     Atom([1,2,3], res_id=1, atom_name="N"),
    ...     Atom([4,5,6], res_id=1, atom_name="CA"),
    ...     Atom([7,8,9], res_id=1, atom_name="C")
    ... ])
    >>> print(atoms)
                1      N                1.000    2.000    3.000
                1      CA               4.000    5.000    6.000
                1      C                7.000    8.000    9.000
    >>> repeat_coord = np.array([
    ...     [[0,0,0], [1,1,1], [2,2,2]],
    ...     [[3,3,3], [4,4,4], [5,5,5]]
    ... ])
    >>> print(repeat(atoms, repeat_coord))
                1      N                0.000    0.000    0.000
                1      CA               1.000    1.000    1.000
                1      C                2.000    2.000    2.000
                1      N                3.000    3.000    3.000
                1      CA               4.000    4.000    4.000
                1      C                5.000    5.000    5.000
    """
    if isinstance(atoms, AtomArray) and coord.ndim != 3:
        raise ValueError(
            f"Expected 3 dimensions for the coordinate array, got {coord.ndim}"
        )
    elif isinstance(atoms, AtomArrayStack) and coord.ndim != 4:
        raise ValueError(
            f"Expected 4 dimensions for the coordinate array, got {coord.ndim}"
        )
    
    repetitions = len(coord)
    orig_length = atoms.array_length()
    new_length = orig_length * repetitions

    if isinstance(atoms, AtomArray):
        if coord.ndim != 3:
            raise ValueError(
                f"Expected 3 dimensions for the coordinate array, "
                f"but got {coord.ndim}"
            )
        repeated = AtomArray(new_length)
        repeated.coord = coord.reshape((new_length, 3))

    elif isinstance(atoms, AtomArrayStack):
        if coord.ndim != 4:
            raise ValueError(
                f"Expected 4 dimensions for the coordinate array, "
                f"but got {coord.ndim}"
            )
        repeated = AtomArrayStack(atoms.stack_depth(), new_length)
        repeated.coord = coord.reshape((atoms.stack_depth(), new_length, 3))
    
    else:
        raise TypeError(
            f"Expected 'AtomArray' or 'AtomArrayStack', "
            f"but got {type(atoms).__name__}"
        )
    
    for category in atoms.get_annotation_categories():
        annot = np.tile(atoms.get_annotation(category), repetitions)
        repeated.set_annotation(category, annot)
    if atoms.bonds is not None:
        bonds = atoms.bonds
        for _ in range(repetitions-1):
            bonds += atoms.bonds
        repeated.bonds = bonds
    if atoms.box is not None:
        repeated.box = atoms.box.copy()
    
    return repeated


def from_template(template, coord, box=None):
    """
    Create an :class:`AtomArrayStack` using template atoms and given
    coordinates.
    
    Parameters
    ----------
    template : AtomArray, shape=(n,) or AtomArrayStack, shape=(m,n)
        The annotation arrays and bonds of the returned stack are taken
        from this template.
    coord : ndarray, dtype=float, shape=(l,n,3)
        The coordinates for each model of the returned stack.
    box : ndarray, optional, dtype=float, shape=(l,3,3)
        The box for each model of the returned stack.
    
    Returns
    -------
    array_stack : AtomArrayStack
        A stack containing the annotation arrays and bonds from
        `template` but the coordinates from `coord` and the boxes from
        `boxes`.
    """
    if template.array_length() != coord.shape[-2]:
        raise ValueError(
            f"Template has {template.array_length()} atoms, but "
            f"{coord.shape[-2]} coordinates are given"
        )

    # Create empty stack with no models
    new_stack = AtomArrayStack(0, template.array_length())
    
    for category in template.get_annotation_categories():
        annot = template.get_annotation(category)
        new_stack.set_annotation(category, annot)
    if template.bonds is not None:
        new_stack.bonds = template.bonds.copy()
    if box is not None:
        new_stack.box = box.copy()
    
    # After setting the coordinates the number of models is the number
    # of models in the new coordinates
    new_stack.coord = coord
    
    return new_stack


def coord(item):
    """
    Get the atom coordinates of the given array.
    
    This may be directly and :class:`Atom`, :class:`AtomArray` or
    :class:`AtomArrayStack` or
    alternatively an (n x 3) or (m x n x 3)  :class:`ndarray`
    containing the coordinates.
    
    Parameters
    ----------
    item : Atom or AtomArray or AtomArrayStack or ndarray
        Returns the :attr:`coord` attribute, if `item` is an
        :class:`Atom`, :class:`AtomArray` or :class:`AtomArrayStack`.
        Directly returns the input, if `item` is a :class:`ndarray`.
    
    Returns
    -------
    coord : ndarray
        Atom coordinates.
    """

    if type(item) in (Atom, AtomArray, AtomArrayStack):
        return item.coord
    elif isinstance(item, np.ndarray):
        return item.astype(np.float32, copy=False)
    else:
        return np.array(item, dtype=np.float32)



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module provides utility for handling data on residue level, rather than
atom level.
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann"
__all__ = ["get_residue_starts", "apply_residue_wise", "spread_residue_wise",
           "get_residue_masks", "get_residue_starts_for",
           "get_residue_positions", "get_residues", "get_residue_count",
           "residue_iter"]

import numpy as np
from .atoms import AtomArray, AtomArrayStack


def get_residue_starts(array, add_exclusive_stop=False):
    """
    Get indices for an atom array, each indicating the beginning of
    a residue.
    
    A new residue starts, either when the chain ID, residue ID,
    insertion code or residue name changes from one to the next atom.
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The atom array (stack) to get the residue starts from.
    add_exclusive_stop : bool, optional
        If true, the exclusive stop of the input atom array, i.e.
        ``array.array_length()``, is added to the returned array of
        start indices as last element.
        
    Returns
    -------
    starts : ndarray, dtype=int
        The start indices of residues in `array`.
    
    Notes
    -----
    This method is internally used by all other residue-related
    functions.

    Examples
    --------

    >>> print(get_residue_starts(atom_array))
    [  0  16  35  56  75  92 116 135 157 169 176 183 197 208 219 226 250 264
     278 292]
    >>> print(get_residue_starts(atom_array, add_exclusive_stop=True))
    [  0  16  35  56  75  92 116 135 157 169 176 183 197 208 219 226 250 264
     278 292 304]
    """
    # These mask are 'true' at indices where the value changes
    chain_id_changes = (array.chain_id[1:] != array.chain_id[:-1])
    res_id_changes   = (array.res_id[1:]   != array.res_id[:-1]  )
    ins_code_changes = (array.ins_code[1:] != array.ins_code[:-1])
    res_name_changes = (array.res_name[1:] != array.res_name[:-1])

    # If any of these annotation arrays change, a new residue starts
    residue_change_mask = (
        chain_id_changes | 
        res_id_changes |
        ins_code_changes | 
        res_name_changes
    )
    
    # Convert mask to indices
    # Add 1, to shift the indices from the end of a residue
    # to the start of a new residue
    residue_starts = np.where(residue_change_mask)[0] +1
    
    # The first residue is not included yet -> Insert '[0]'
    if add_exclusive_stop:
        return np.concatenate(([0], residue_starts, [array.array_length()]))
    else:
        return np.concatenate(([0], residue_starts))


def apply_residue_wise(array, data, function, axis=None):
    """
    Apply a function to intervals of data, where each interval
    corresponds to one residue.
    
    The function takes an atom array (stack) and an data array
    (`ndarray`) of the same length. The function iterates through the
    residue IDs of the atom array (stack) and identifies intervals of
    the same ID. Then the data is
    partitioned into the same intervals, and each interval (also an
    :class:`ndarray`) is put as parameter into `function`. Each return value is
    stored as element in the resulting :class:`ndarray`, therefore each element
    corresponds to one residue. 
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The atom array (stack) to determine the residues from.
    data : ndarray
        The data, whose intervals are the parameter for `function`. Must
        have same length as `array`.
    function : function
        The `function` must have either the form *f(data)* or
        *f(data, axis)* in case `axis` is given. Every `function` call
        must return a value with the same shape and data type.
    axis : int, optional
        This value is given to the `axis` parameter of `function`.
        
    Returns
    -------
    processed_data : ndarray
        Residue-wise evaluation of `data` by `function`. The size of the
        first dimension of this array is equal to the amount of
        residues.
        
    Examples
    --------
    Calculate residue-wise SASA from atom-wise SASA of a 20 residue
    peptide.
    
    >>> sasa_per_atom = sasa(atom_array)
    >>> print(len(sasa_per_atom))
    304
    >>> sasa_per_residue = apply_residue_wise(atom_array, sasa_per_atom, np.nansum)
    >>> print(len(sasa_per_residue))
    20
    >>> print(sasa_per_residue)
    [157.979 117.136  94.983 115.485 113.583  23.471  93.013 144.173  61.561
      38.885   0.792 114.053 108.568  27.888  83.583 113.016 114.318  74.281
      47.811 172.035]

    Calculate the centroids of each residue for the same peptide.
        
    >>> print(len(atom_array))
    304
    >>> centroids = apply_residue_wise(atom_array, atom_array.coord,
    ...                                np.average, axis=0)
    >>> print(len(centroids))
    20
    >>> print(centroids)
    [[-9.582  3.378 -2.073]
     [-4.670  5.816 -1.860]
     [-2.461  3.060  3.076]
     [-7.211 -0.396  1.013]
     [-4.698 -1.080 -4.284]
     [ 1.172  0.206  1.038]
     [-2.160 -2.245  3.541]
     [-3.682 -5.540 -2.895]
     [ 0.711 -5.409 -2.549]
     [ 2.002 -6.322  1.695]
     [ 2.799 -3.140  2.327]
     [ 5.901 -2.489  4.845]
     [ 6.754 -6.712  3.094]
     [ 5.699 -5.101 -1.209]
     [ 9.295 -2.970 -1.835]
     [ 5.518 -1.521 -3.473]
     [ 7.219  3.673 -0.684]
     [ 4.007  4.364  2.674]
     [ 0.341  5.575 -0.254]
     [ 1.194 10.416  1.130]]
    """
    starts = get_residue_starts(array, add_exclusive_stop=True)
    # The result array
    processed_data = None
    for i in range(len(starts)-1):
        interval = data[starts[i]:starts[i+1]]
        if axis == None:
            value = function(interval)
        else:
            value = function(interval, axis=axis)
        value = function(interval, axis=axis)
        # Identify the shape of the resulting array by evaluation
        # of the function return value for the first interval
        if processed_data is None:
            if isinstance(value, np.ndarray):
                # Maximum length of the processed data
                # is length of interval of size 1 -> length of all IDs
                # (equal to atom array length)
                processed_data = np.zeros((len(starts)-1,) + value.shape,
                                          dtype=value.dtype)
            else:
                # Scalar value -> one dimensional result array
                processed_data = np.zeros(len(starts)-1,
                                          dtype=type(value))
        # Write values into result arrays
        processed_data[i] = value
    return processed_data


def spread_residue_wise(array, input_data):
    """
    Expand residue-wise data to atom-wise data.

    Each value in the residue-wise input is assigned to all atoms of
    this residue:
    
    ``output_data[i] = input_data[j]``,
    *i* is incremented from atom to atom,
    *j* is incremented every residue change.
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The atom array (stack) to determine the residues from.
    input_data : ndarray
        The data to be spread. The length of axis=0 must be equal to
        the amount of different residue IDs in `array`.
        
    Returns
    -------
    output_data : ndarray
        Residue-wise spread `input_data`. Length is the same as
        `array_length()` of `array`.
        
    Examples
    --------
    Spread secondary structure annotation to every atom of a 20 residue
    peptide (with 304 atoms).
    
        >>> sse = annotate_sse(atom_array, "A")
        >>> print(len(sse))
        20
        >>> print(sse)
        ['c' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c'
         'c' 'c']
        >>> atom_wise_sse = spread_residue_wise(atom_array, sse)
        >>> print(len(atom_wise_sse))
        304
        >>> print(atom_wise_sse)
        ['c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'a' 'a'
         'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a'
         'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a'
         'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a'
         'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a'
         'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a'
         'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a'
         'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a'
         'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a' 'a'
         'a' 'a' 'a' 'a' 'a' 'a' 'a' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c'
         'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c'
         'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c'
         'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c'
         'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c'
         'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c'
         'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c'
         'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c' 'c']
    """
    starts = get_residue_starts(array)
    output_data = np.zeros(array.array_length(), dtype=input_data.dtype)
    for i in range(len(starts)-1):
        output_data[starts[i]:starts[i+1]] = input_data[i]
    output_data[starts[-1]:] = input_data[-1]
    return output_data


def get_residue_masks(array, indices):
    """
    Get boolean masks indicating the residues to which the given atom
    indices belong.

    Parameters
    ----------
    array : AtomArray, shape=(n,) or AtomArrayStack, shape=(m,n)
        The atom array (stack) to determine the residues from.
    indices : ndarray, dtype=int, shape=(k,)
        These indices indicate the atoms to get the corresponding
        residues for.
        Negative indices are not allowed.
    
    Returns
    -------
    residues_masks : ndarray, dtype=bool, shape=(k,n)
        Multiple boolean masks, one for each given index in `indices`.
        Each array masks the atoms that belong to the same residue as
        the atom at the given index.
    
    See also
    --------
    get_residue_starts_for
    get_residue_positions
    
    Examples
    --------

    >>> indices = [5, 42]
    >>> residue_masks = get_residue_masks(atom_array, indices)
    >>> print(atom_array[indices[0]])
        A       1  ASN CG     C       -10.915    3.130   -2.611
    >>> print(atom_array[residue_masks[0]])
        A       1  ASN N      N        -8.901    4.127   -0.555
        A       1  ASN CA     C        -8.608    3.135   -1.618
        A       1  ASN C      C        -7.117    2.964   -1.897
        A       1  ASN O      O        -6.634    1.849   -1.758
        A       1  ASN CB     C        -9.437    3.396   -2.889
        A       1  ASN CG     C       -10.915    3.130   -2.611
        A       1  ASN OD1    O       -11.269    2.700   -1.524
        A       1  ASN ND2    N       -11.806    3.406   -3.543
        A       1  ASN H1     H        -8.330    3.957    0.261
        A       1  ASN H2     H        -8.740    5.068   -0.889
        A       1  ASN H3     H        -9.877    4.041   -0.293
        A       1  ASN HA     H        -8.930    2.162   -1.239
        A       1  ASN HB2    H        -9.310    4.417   -3.193
        A       1  ASN HB3    H        -9.108    2.719   -3.679
        A       1  ASN HD21   H       -11.572    3.791   -4.444
        A       1  ASN HD22   H       -12.757    3.183   -3.294
    >>> print(atom_array[indices[1]])
        A       3  TYR CD2    C        -1.820    4.326    3.332
    >>> print(atom_array[residue_masks[1]])
        A       3  TYR N      N        -4.354    3.455   -0.111
        A       3  TYR CA     C        -3.690    2.738    0.981
        A       3  TYR C      C        -4.102    1.256    1.074
        A       3  TYR O      O        -3.291    0.409    1.442
        A       3  TYR CB     C        -3.964    3.472    2.302
        A       3  TYR CG     C        -2.824    3.339    3.290
        A       3  TYR CD1    C        -2.746    2.217    4.138
        A       3  TYR CD2    C        -1.820    4.326    3.332
        A       3  TYR CE1    C        -1.657    2.076    5.018
        A       3  TYR CE2    C        -0.725    4.185    4.205
        A       3  TYR CZ     C        -0.639    3.053    5.043
        A       3  TYR OH     O         0.433    2.881    5.861
        A       3  TYR H      H        -4.934    4.245    0.120
        A       3  TYR HA     H        -2.615    2.768    0.796
        A       3  TYR HB2    H        -4.117    4.513    2.091
        A       3  TYR HB3    H        -4.886    3.096    2.750
        A       3  TYR HD1    H        -3.513    1.456    4.101
        A       3  TYR HD2    H        -1.877    5.200    2.695
        A       3  TYR HE1    H        -1.576    1.221    5.669
        A       3  TYR HE2    H         0.033    4.952    4.233
        A       3  TYR HH     H         1.187    3.395    5.567
    """
    indices = np.asarray(indices)
    starts = get_residue_starts(array, add_exclusive_stop=True)
    masks = np.zeros((len(indices), array.array_length()), dtype=bool)

    if (indices < 0).any():
        raise ValueError("This function does not support negative indices")
    if (indices >= array.array_length()).any():
        index = np.min(np.where(indices >= array.array_length())[0])
        raise ValueError(
            f"Index {index} is out of range for "
            f"an atom array with length {array.array_length()}"
        )
    
    insertion_points = np.searchsorted(starts, indices, side="right") - 1
    for i, point in enumerate(insertion_points):
        masks[i, starts[point] : starts[point+1]] = True
    
    return masks


def get_residue_starts_for(array, indices):
    """
    For each given atom index, get the index that points to the
    start of the residue that atom belongs to.

    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The atom array (stack) to determine the residues from.
    indices : ndarray, dtype=int, shape=(k,)
        These indices point to the atoms to get the corresponding
        residue starts for.
        Negative indices are not allowed.
    
    Returns
    -------
    start_indices : ndarray, dtype=int, shape=(k,)
        The indices that point to the residue starts for the input
        `indices`.
    
    See also
    --------
    get_residue_masks
    get_residue_positions
    
    Examples
    --------

    >>> indices = [5, 42]
    >>> residue_starts = get_residue_starts_for(atom_array, indices)
    >>> print(residue_starts)
    [ 0 35]
    >>> print(atom_array[indices[0]])
        A       1  ASN CG     C       -10.915    3.130   -2.611
    >>> print(atom_array[residue_starts[0]])
        A       1  ASN N      N        -8.901    4.127   -0.555
    >>> print(atom_array[indices[1]])
        A       3  TYR CD2    C        -1.820    4.326    3.332
    >>> print(atom_array[residue_starts[1]])
        A       3  TYR N      N        -4.354    3.455   -0.111
    """
    indices = np.asarray(indices)
    starts = get_residue_starts(array)

    if (indices < 0).any():
        raise ValueError("This function does not support negative indices")
    if (indices >= array.array_length()).any():
        index = np.min(np.where(indices >= array.array_length())[0])
        raise ValueError(
            f"Index {index} is out of range for "
            f"an atom array with length {array.array_length()}"
        )
    
    insertion_points = np.searchsorted(starts, indices, side="right") - 1
    return starts[insertion_points]


def get_residue_positions(array, indices):
    """
    For each given atom index, obtain the position of the residue
    corresponding to this index in the input `array`.

    For example, the position of the first residue in the atom array is
    ``0``, the the position of the second residue is ``1``, etc.

    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The atom array (stack) to determine the residues from.
    indices : ndarray, dtype=int, shape=(k,)
        These indices point to the atoms to get the corresponding
        residue positions for.
        Negative indices are not allowed.
    
    Returns
    -------
    start_indices : ndarray, dtype=int, shape=(k,)
        The indices that point to the position of the residues.
    
    See also
    --------
    get_residue_masks
    get_residue_starts_for

    Examples
    --------
    >>> atom_index = [5, 42]
    >>> print(atom_array.res_name[atom_index])
    ['ASN' 'TYR']
    >>> _, residues = get_residues(atom_array)
    >>> print(residues)
    ['ASN' 'LEU' 'TYR' 'ILE' 'GLN' 'TRP' 'LEU' 'LYS' 'ASP' 'GLY' 'GLY' 'PRO'
     'SER' 'SER' 'GLY' 'ARG' 'PRO' 'PRO' 'PRO' 'SER']
    >>> residue_index = get_residue_positions(atom_array, atom_index)
    >>> print(residue_index)
    [0 2]
    >>> print(residues[residue_index])
    ['ASN' 'TYR']
    """
    indices = np.asarray(indices)
    starts = get_residue_starts(array)

    if (indices < 0).any():
        raise ValueError("This function does not support negative indices")
    if (indices >= array.array_length()).any():
        index = np.min(np.where(indices >= array.array_length())[0])
        raise ValueError(
            f"Index {index} is out of range for "
            f"an atom array with length {array.array_length()}"
        )
    
    return np.searchsorted(starts, indices, side="right") - 1


def get_residues(array):
    """
    Get the residue IDs and names of an atom array (stack).
    
    The residues are listed in the same order they occur in the array
    (stack).
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The atom array (stack) to determine the residues from.
        
    Returns
    -------
    ids : ndarray, dtype=int
        List of residue IDs.
    names : ndarray, dtype="U3"
        List of residue names.
        
    Examples
    --------
    Get the residue names of a 20 residue peptide.
    
    >>> print(atom_array.res_name)
    ['ASN' 'ASN' 'ASN' 'ASN' 'ASN' 'ASN' 'ASN' 'ASN' 'ASN' 'ASN' 'ASN' 'ASN'
     'ASN' 'ASN' 'ASN' 'ASN' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU'
     'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'TYR'
     'TYR' 'TYR' 'TYR' 'TYR' 'TYR' 'TYR' 'TYR' 'TYR' 'TYR' 'TYR' 'TYR' 'TYR'
     'TYR' 'TYR' 'TYR' 'TYR' 'TYR' 'TYR' 'TYR' 'TYR' 'ILE' 'ILE' 'ILE' 'ILE'
     'ILE' 'ILE' 'ILE' 'ILE' 'ILE' 'ILE' 'ILE' 'ILE' 'ILE' 'ILE' 'ILE' 'ILE'
     'ILE' 'ILE' 'ILE' 'GLN' 'GLN' 'GLN' 'GLN' 'GLN' 'GLN' 'GLN' 'GLN' 'GLN'
     'GLN' 'GLN' 'GLN' 'GLN' 'GLN' 'GLN' 'GLN' 'GLN' 'TRP' 'TRP' 'TRP' 'TRP'
     'TRP' 'TRP' 'TRP' 'TRP' 'TRP' 'TRP' 'TRP' 'TRP' 'TRP' 'TRP' 'TRP' 'TRP'
     'TRP' 'TRP' 'TRP' 'TRP' 'TRP' 'TRP' 'TRP' 'TRP' 'LEU' 'LEU' 'LEU' 'LEU'
     'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU' 'LEU'
     'LEU' 'LEU' 'LEU' 'LYS' 'LYS' 'LYS' 'LYS' 'LYS' 'LYS' 'LYS' 'LYS' 'LYS'
     'LYS' 'LYS' 'LYS' 'LYS' 'LYS' 'LYS' 'LYS' 'LYS' 'LYS' 'LYS' 'LYS' 'LYS'
     'LYS' 'ASP' 'ASP' 'ASP' 'ASP' 'ASP' 'ASP' 'ASP' 'ASP' 'ASP' 'ASP' 'ASP'
     'ASP' 'GLY' 'GLY' 'GLY' 'GLY' 'GLY' 'GLY' 'GLY' 'GLY' 'GLY' 'GLY' 'GLY'
     'GLY' 'GLY' 'GLY' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO'
     'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'SER' 'SER' 'SER' 'SER' 'SER' 'SER' 'SER'
     'SER' 'SER' 'SER' 'SER' 'SER' 'SER' 'SER' 'SER' 'SER' 'SER' 'SER' 'SER'
     'SER' 'SER' 'SER' 'GLY' 'GLY' 'GLY' 'GLY' 'GLY' 'GLY' 'GLY' 'ARG' 'ARG'
     'ARG' 'ARG' 'ARG' 'ARG' 'ARG' 'ARG' 'ARG' 'ARG' 'ARG' 'ARG' 'ARG' 'ARG'
     'ARG' 'ARG' 'ARG' 'ARG' 'ARG' 'ARG' 'ARG' 'ARG' 'ARG' 'ARG' 'PRO' 'PRO'
     'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO'
     'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO'
     'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO' 'PRO'
     'PRO' 'PRO' 'PRO' 'PRO' 'SER' 'SER' 'SER' 'SER' 'SER' 'SER' 'SER' 'SER'
     'SER' 'SER' 'SER' 'SER']
    >>> ids, names = get_residues(atom_array)
    >>> print(names)
    ['ASN' 'LEU' 'TYR' 'ILE' 'GLN' 'TRP' 'LEU' 'LYS' 'ASP' 'GLY' 'GLY' 'PRO'
     'SER' 'SER' 'GLY' 'ARG' 'PRO' 'PRO' 'PRO' 'SER']
    """
    starts = get_residue_starts(array)
    return array.res_id[starts], array.res_name[starts]


def get_residue_count(array):
    """
    Get the amount of residues in an atom array (stack).
    
    The count is determined from the `res_id` and `chain_id` annotation.
    Each time the residue ID or chain ID changes,
    the count is incremented. Special rules apply to hetero residues.
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The atom array (stack) to determine the residues from.
        
    Returns
    -------
    count : int
        Amount of residues.
    """
    return len(get_residue_starts(array))


def residue_iter(array):
    """
    Iterate over all residues in an atom array (stack).
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The atom array (stack) to iterate over.
        
    Yields
    ------
    residue : AtomArray or AtomArrayStack
        A single residue of the input `array`.
    
    Examples
    --------

    >>> for res in residue_iter(atom_array[:50]):
    ...     print("New residue")
    ...     print(res)
    ...     print()
    New residue
        A       1  ASN N      N        -8.901    4.127   -0.555
        A       1  ASN CA     C        -8.608    3.135   -1.618
        A       1  ASN C      C        -7.117    2.964   -1.897
        A       1  ASN O      O        -6.634    1.849   -1.758
        A       1  ASN CB     C        -9.437    3.396   -2.889
        A       1  ASN CG     C       -10.915    3.130   -2.611
        A       1  ASN OD1    O       -11.269    2.700   -1.524
        A       1  ASN ND2    N       -11.806    3.406   -3.543
        A       1  ASN H1     H        -8.330    3.957    0.261
        A       1  ASN H2     H        -8.740    5.068   -0.889
        A       1  ASN H3     H        -9.877    4.041   -0.293
        A       1  ASN HA     H        -8.930    2.162   -1.239
        A       1  ASN HB2    H        -9.310    4.417   -3.193
        A       1  ASN HB3    H        -9.108    2.719   -3.679
        A       1  ASN HD21   H       -11.572    3.791   -4.444
        A       1  ASN HD22   H       -12.757    3.183   -3.294
    <BLANKLINE>
    New residue
        A       2  LEU N      N        -6.379    4.031   -2.228
        A       2  LEU CA     C        -4.923    4.002   -2.452
        A       2  LEU C      C        -4.136    3.187   -1.404
        A       2  LEU O      O        -3.391    2.274   -1.760
        A       2  LEU CB     C        -4.411    5.450   -2.619
        A       2  LEU CG     C        -4.795    6.450   -1.495
        A       2  LEU CD1    C        -3.612    6.803   -0.599
        A       2  LEU CD2    C        -5.351    7.748   -2.084
        A       2  LEU H      H        -6.821    4.923   -2.394
        A       2  LEU HA     H        -4.750    3.494   -3.403
        A       2  LEU HB2    H        -3.340    5.414   -2.672
        A       2  LEU HB3    H        -4.813    5.817   -3.564
        A       2  LEU HG     H        -5.568    6.022   -0.858
        A       2  LEU HD11   H        -3.207    5.905   -0.146
        A       2  LEU HD12   H        -2.841    7.304   -1.183
        A       2  LEU HD13   H        -3.929    7.477    0.197
        A       2  LEU HD21   H        -4.607    8.209   -2.736
        A       2  LEU HD22   H        -6.255    7.544   -2.657
        A       2  LEU HD23   H        -5.592    8.445   -1.281
    <BLANKLINE>
    New residue
        A       3  TYR N      N        -4.354    3.455   -0.111
        A       3  TYR CA     C        -3.690    2.738    0.981
        A       3  TYR C      C        -4.102    1.256    1.074
        A       3  TYR O      O        -3.291    0.409    1.442
        A       3  TYR CB     C        -3.964    3.472    2.302
        A       3  TYR CG     C        -2.824    3.339    3.290
        A       3  TYR CD1    C        -2.746    2.217    4.138
        A       3  TYR CD2    C        -1.820    4.326    3.332
        A       3  TYR CE1    C        -1.657    2.076    5.018
        A       3  TYR CE2    C        -0.725    4.185    4.205
        A       3  TYR CZ     C        -0.639    3.053    5.043
        A       3  TYR OH     O         0.433    2.881    5.861
        A       3  TYR H      H        -4.934    4.245    0.120
        A       3  TYR HA     H        -2.615    2.768    0.796
        A       3  TYR HB2    H        -4.117    4.513    2.091
    <BLANKLINE>
    """
    # The exclusive stop is appended to the residue starts
    starts = np.append(get_residue_starts(array), [array.array_length()])
    for i in range(len(starts)-1):
        yield array[..., starts[i] : starts[i+1]]



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from tempfile import TemporaryFile
import itertools
import glob
from os.path import join, splitext
import pytest
from pytest import approx
import numpy as np
import biotite
import biotite.structure as struc
import biotite.structure.io.pdb as pdb
import biotite.structure.io.pdb.hybrid36 as hybrid36
import biotite.structure.io.pdbx as pdbx
import biotite.structure.io as io
from ..util import data_dir


def test_get_model_count():
    pdb_file = pdb.PDBFile.read(join(data_dir("structure"), "1l2y.pdb"))
    # Test also the thin wrapper around the method
    # 'get_model_count()'
    test_model_count = pdb.get_model_count(pdb_file)
    ref_model_count = pdb.get_structure(pdb_file).stack_depth()
    assert test_model_count == ref_model_count


@pytest.mark.parametrize(
    "path, model, hybrid36, include_bonds",
    itertools.product(
        glob.glob(join(data_dir("structure"), "*.pdb")),
        [None, 1, -1],
        [False, True],
        [False, True]
    )
)
def test_array_conversion(path, model, hybrid36, include_bonds):
    pdb_file = pdb.PDBFile.read(path)
    # Test also the thin wrapper around the methods
    # 'get_structure()' and 'set_structure()'
    try:
        array1 = pdb.get_structure(
            pdb_file, model=model, include_bonds=include_bonds
        )
    except biotite.InvalidFileError:
        if model is None:
            # The file cannot be parsed into an AtomArrayStack,
            # as the models contain different numbers of atoms
            # -> skip this test case
            return
        else:
            raise
    
    if hybrid36 and (array1.res_id < 1).any():
        with pytest.raises(
            ValueError,
            match="Only positive integers can be converted "
                  "into hybrid-36 notation"
        ):
            pdb_file = pdb.PDBFile()
            pdb.set_structure(pdb_file, array1, hybrid36=hybrid36)
        return
    else:
        pdb_file = pdb.PDBFile()
        pdb.set_structure(pdb_file, array1, hybrid36=hybrid36)
    
    array2 = pdb.get_structure(
        pdb_file, model=model, include_bonds=include_bonds
    )
    
    if array1.box is not None:
        assert np.allclose(array1.box, array2.box)
    assert array1.bonds == array2.bonds
    for category in array1.get_annotation_categories():
        assert array1.get_annotation(category).tolist() == \
               array2.get_annotation(category).tolist()
    assert array1.coord.tolist() == array2.coord.tolist()


@pytest.mark.parametrize(
    "path, model",
    itertools.product(
        glob.glob(join(data_dir("structure"), "*.pdb")),
        [None, 1, -1]
    )
)
def test_pdbx_consistency(path, model):
    cif_path = splitext(path)[0] + ".cif"
    pdb_file = pdb.PDBFile.read(path)
    try:
        a1 = pdb_file.get_structure(model=model)
    except biotite.InvalidFileError:
        if model is None:
            # The file cannot be parsed into an AtomArrayStack,
            # as the models contain different numbers of atoms
            # -> skip this test case
            return
        else:
            raise

    pdbx_file = pdbx.PDBxFile.read(cif_path)
    a2 = pdbx.get_structure(pdbx_file, model=model)
    
    if a2.box is not None:
        assert np.allclose(a1.box, a2.box)
    assert a1.bonds == a2.bonds
    for category in a1.get_annotation_categories():
        assert a1.get_annotation(category).tolist() == \
               a2.get_annotation(category).tolist()
    assert a1.coord.tolist() == a2.coord.tolist()


@pytest.mark.parametrize("hybrid36", [False, True])
def test_extra_fields(hybrid36):
    path = join(data_dir("structure"), "1l2y.pdb")
    pdb_file = pdb.PDBFile.read(path)
    stack1 = pdb_file.get_structure(
        extra_fields=[
            "atom_id", "b_factor", "occupancy", "charge"
        ]
    )

    with pytest.raises(ValueError):
        pdb_file.get_structure(extra_fields=["unsupported_field"])
    
    # Add non-neutral charge values,
    # as the input PDB has only neutral charges
    stack1.charge[0] = -1
    stack1.charge[1] = 2

    pdb_file = pdb.PDBFile()
    pdb_file.set_structure(stack1, hybrid36=hybrid36)
    
    stack2 = pdb_file.get_structure(
        extra_fields=[
            "atom_id", "b_factor", "occupancy", "charge"
        ]
    )
    
    assert stack1.ins_code.tolist() == stack2.ins_code.tolist()
    assert stack1.atom_id.tolist() == stack2.atom_id.tolist()
    assert stack1.b_factor.tolist() == approx(stack2.b_factor.tolist())
    assert stack1.occupancy.tolist() == approx(stack2.occupancy.tolist())
    assert stack1.charge.tolist() == stack2.charge.tolist()
    assert stack1 == stack2


@pytest.mark.filterwarnings("ignore")
def test_guess_elements():
    # Read valid pdb file
    path = join(data_dir("structure"), "1l2y.pdb")
    pdb_file = pdb.PDBFile.read(path)
    stack = pdb_file.get_structure()

    # Remove all elements
    removed_stack = stack.copy()
    removed_stack.element[:] = ''

    # Save stack without elements to tmp file
    temp = TemporaryFile("w+")
    tmp_pdb_file = pdb.PDBFile()
    tmp_pdb_file.set_structure(removed_stack)
    tmp_pdb_file.write(temp)

    # Read new stack from file with guessed elements
    temp.seek(0)
    guessed_pdb_file = pdb.PDBFile.read(temp)
    temp.close()
    guessed_stack = guessed_pdb_file.get_structure()

    assert guessed_stack.element.tolist() == stack.element.tolist()


@pytest.mark.parametrize(
    "path, model",
    itertools.product(
        glob.glob(join(data_dir("structure"), "*.pdb")),
        [None, 1, -1]
    )
)
def test_box_shape(path, model):
    pdb_file = pdb.PDBFile.read(path)
    try:
        a = pdb_file.get_structure(model=model)
    except biotite.InvalidFileError:
        if model is None:
            # The file cannot be parsed into an AtomArrayStack,
            # as the models contain different numbers of atoms
            # -> skip this test case
            return
        else:
            raise

    if isinstance(a, struc.AtomArray):
        expected_box_dim = (3, 3)
    else:
        expected_box_dim = (len(a), 3, 3)
    assert expected_box_dim == a.box.shape


def test_box_parsing():
    path = join(data_dir("structure"), "1igy.pdb")
    pdb_file = pdb.PDBFile.read(path)
    a = pdb_file.get_structure()
    expected_box = np.array([[
        [66.65,   0.00, 0.00],
        [0.00,  190.66, 0.00],
        [-24.59,  0.00, 68.84]
    ]])

    assert expected_box.flatten().tolist() \
           == approx(a.box.flatten().tolist(), abs=1e-2)


def test_id_overflow():
    # Create an atom array >= 100k atoms
    length = 100000
    a = struc.AtomArray(length)
    a.coord = np.zeros(a.coord.shape)
    a.chain_id = np.full(length, "A")
    # Create residue IDs over 10000
    a.res_id = np.arange(1, length+1)
    a.res_name = np.full(length, "GLY")
    a.hetero = np.full(length, False)
    a.atom_name = np.full(length, "CA")
    a.element = np.full(length, "C")
    
    # Write stack to pdb file and make sure a warning is thrown
    with pytest.warns(UserWarning):
        temp = TemporaryFile("w+")
        pdb_file = pdb.PDBFile()
        pdb_file.set_structure(a)
        pdb_file.write(temp)

    # Assert file can be read properly
    temp.seek(0)
    a2 = pdb.get_structure(pdb.PDBFile.read(temp))
    assert(a2.array_length() == a.array_length())
    
    # Manually check if the written atom id is correct
    temp.seek(0)
    last_line = temp.readlines()[-1]
    atom_id = int(last_line.split()[1])
    assert(atom_id == 1)

    temp.close()
    
    # Write stack as hybrid-36 pdb file: no warning should be thrown
    with pytest.warns(None) as record:
        temp = TemporaryFile("w+")
        tmp_pdb_file = pdb.PDBFile()
        tmp_pdb_file.set_structure(a, hybrid36=True)
        tmp_pdb_file.write(temp)
    assert len(record) == 0

    # Manually check if the output is written as correct hybrid-36
    temp.seek(0)
    last_line = temp.readlines()[-1]
    atom_id = last_line.split()[1]
    assert(atom_id == "A0000")
    res_id = last_line.split()[4][1:]
    assert(res_id == "BXG0")

    temp.close()


@pytest.mark.parametrize("model", [None, 1, 10])
def test_get_coord(model):
    # Choose a structure without inscodes and altlocs
    # to avoid atom filtering in reference atom array (stack)
    path = join(data_dir("structure"), "1l2y.pdb")
    pdb_file = pdb.PDBFile.read(path)
    
    try:
        ref_coord = pdb_file.get_structure(model=model).coord
    except biotite.InvalidFileError:
        if model is None:
            # The file cannot be parsed into an AtomArrayStack,
            # as the models contain different numbers of atoms
            # -> skip this test case
            return
        else:
            raise
    
    test_coord = pdb_file.get_coord(model=model)
    
    assert test_coord.shape == ref_coord.shape
    assert (test_coord == ref_coord).all()


np.random.seed(0)
N = 200
LENGTHS = [3, 4, 5]
@pytest.mark.parametrize(
    "number, length",
    zip(
        list(itertools.chain(*[
            np.random.randint(0, hybrid36.max_hybrid36_number(length), N)
            for length in LENGTHS
        ])),
        list(itertools.chain(*[
            [length] * N for length in LENGTHS
        ]))
    )
)
def test_hybrid36_codec(number, length):
    string = hybrid36.encode_hybrid36(number, length)
    test_number = hybrid36.decode_hybrid36(string)
    assert test_number == number


def test_max_hybrid36_number():
    assert hybrid36.max_hybrid36_number(4) == 2436111
    assert hybrid36.max_hybrid36_number(5) == 87440031


def test_bond_parsing():
    """
    Compare parsing of bonds from PDB with output from
    :func:`connect_via_residue_names()`.
    """
    # Choose a structure with CONECT records to test these as well
    path = join(data_dir("structure"), "3o5r.pdb")
    pdb_file = pdb.PDBFile.read(path)
    atoms = pdb.get_structure(pdb_file, model=1, include_bonds=True)
    
    test_bonds = atoms.bonds

    ref_bonds = struc.connect_via_residue_names(atoms)
    ref_bonds.remove_bond_order()

    assert test_bonds.as_set() == ref_bonds.as_set()


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import numpy as np
import pytest
import biotite.sequence as seq
import biotite.sequence.align as align
from biotite.application.tantan import TantanApp
from ..util import is_not_installed

@pytest.fixture
def simple_matrix():
    alph = seq.NucleotideSequence.alphabet_unamb
    return align.SubstitutionMatrix(
        alph, alph, np.array(
            [[ 1, -1, -1, -1],
             [-1,  1, -1, -1],
             [-1, -1,  1, -1],
             [-1, -1, -1,  1]]
        )
    )


@pytest.mark.skipif(
    is_not_installed("tantan"), reason="tantan is not installed"
)
@pytest.mark.parametrize("use_custom_matrix", [False, True])
def test_nucleotide(simple_matrix, use_custom_matrix):
    """
    Test masking a nucleotide sequence based on a known example.
    """
    seq_string = "TGCAAGCTATTAGGCTTAGGTCAGTGCttaagcttaggtcagtgcAACATA"
    sequence = seq.NucleotideSequence(seq_string)

    if use_custom_matrix:
        matrix = simple_matrix
    else:
        matrix = None

    test_mask = TantanApp.mask_repeats(sequence, matrix)

    ref_mask = [True if char.islower() else False for char in seq_string]

    assert len(test_mask) == len(ref_mask)
    assert np.all(test_mask.tolist() == ref_mask)


@pytest.mark.skipif(
    is_not_installed("tantan"), reason="tantan is not installed"
)
@pytest.mark.parametrize("use_custom_matrix", [False, True])
def test_protein(use_custom_matrix):
    """
    Test masking a protein sequence based on a known example.
    """
    seq_string = "MAPKINASekinasekinase"
    sequence = seq.ProteinSequence(seq_string)

    if use_custom_matrix:
        matrix = align.SubstitutionMatrix.std_protein_matrix()
    else:
        matrix = None

    test_mask = TantanApp.mask_repeats(sequence, matrix)

    ref_mask = [True if char.islower() else False for char in seq_string]

    assert len(test_mask) == len(ref_mask)
    assert np.all(test_mask.tolist() == ref_mask)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import biotite.structure as struc
import biotite.structure.io.npz as npz
import numpy as np
from os.path import join
from ..util import data_dir
import pytest


@pytest.fixture
def sample_array():
    file = npz.NpzFile.read(join(data_dir("structure"), "1l2y.npz"))
    return file.get_structure()[0]

@pytest.fixture
def gapped_sample_array(sample_array):
    atom_ids = np.arange(1, sample_array.shape[0]+1)
    sample_array.add_annotation("atom_id", dtype=int)
    sample_array.atom_id = atom_ids
    sample_array = sample_array[sample_array.res_id != 5]
    sample_array = sample_array[(sample_array.res_id != 9) |
                                (sample_array.atom_name != "N")]
    return sample_array

@pytest.fixture
def duplicate_sample_array(sample_array):
    sample_array[42] = sample_array[10]
    sample_array[234] = sample_array[123]
    return sample_array

def test_atom_id_continuity_check(gapped_sample_array):
    discon = struc.check_atom_id_continuity(gapped_sample_array)
    discon_array = gapped_sample_array[discon]
    assert discon_array.atom_id.tolist() == [93, 159]

def test_res_id_continuity_check(gapped_sample_array):
    discon = struc.check_res_id_continuity(gapped_sample_array)
    discon_array = gapped_sample_array[discon]
    assert discon_array.res_id.tolist() == [6]

def test_bond_continuity_check(gapped_sample_array):
    discon = struc.check_bond_continuity(gapped_sample_array)
    discon_array = gapped_sample_array[discon]
    assert discon_array.res_id.tolist() == [6,9]

def test_duplicate_atoms_check(duplicate_sample_array):
    discon = struc.check_duplicate_atoms(duplicate_sample_array)
    assert discon.tolist() == [42,234]

def test_renum_res_ids(gapped_sample_array): 
    renumbered_array = struc.renumber_res_ids(gapped_sample_array)
    # if renumbering was successful, this should not raise
    with pytest.raises(AssertionError):
        test_res_id_continuity_check(renumbered_array)

def test_renum_atom_ids(gapped_sample_array):
    renumbered_array = struc.renumber_atom_ids(gapped_sample_array)
    # if renumbering was successful, this should not raise
    with pytest.raises(AssertionError):
        test_atom_id_continuity_check(renumbered_array)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module provides different transformations#
that can be applied on structures.
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann", "Claude J. Rogers"
__all__ = ["translate", "rotate", "rotate_centered", "rotate_about_axis",
           "orient_principal_components", "align_vectors"]

import numpy as np
from .geometry import centroid
from .error import BadStructureError
from .atoms import Atom, AtomArray, AtomArrayStack, coord
from .util import norm_vector, vector_dot, matrix_rotate


def translate(atoms, vector):
    """
    Translate the given atoms or coordinates by a given vector.
    
    Parameters
    ----------
    atoms : Atom or AtomArray or AtomArrayStack or ndarray, shape=(3,) or shape=(n,3) or shape=(m,n,3)
        The atoms of which the coordinates are transformed.
        The coordinates can be directly provided as :class:`ndarray`.
    vector: array-like, shape=(3,) or shape=(n,3) or shape=(m,n,3)
        The translation vector :math:`(x, y, z)`.
    
    Returns
    -------
    transformed : Atom or AtomArray or AtomArrayStack or ndarray, shape=(3,) or shape=(n,3) or shape=(m,n,3)
        A copy of the input atoms or coordinates, translated by the
        given vector.
    """
    positions = coord(atoms).copy()
    vector = np.asarray(vector)
    
    if vector.shape[-1] != 3:
        raise ValueError("Translation vector must contain 3 coordinates")
    positions += vector
    return _put_back(atoms, positions)


def rotate(atoms, angles):
    """
    Rotate the given atoms or coordinates about the *x*, *y* and *z*
    axes by given angles.
    
    The rotations are centered at the origin and are performed
    sequentially in the order *x*, *y*, *z*.
    
    Parameters
    ----------
    atoms : Atom or AtomArray or AtomArrayStack or ndarray, shape=(3,) or shape=(n,3) or shape=(m,n,3)
        The atoms of which the coordinates are transformed.
        The coordinates can be directly provided as :class:`ndarray`.
    angles: array-like, length=3
        The rotation angles in radians around *x*, *y* and *z*.
    
    Returns
    -------
    transformed : Atom or AtomArray or AtomArrayStack or ndarray, shape=(3,) or shape=(n,3) or shape=(m,n,3)
        A copy of the input atoms or coordinates, rotated by the given
        angles.
        
    See Also
    --------
    rotate_centered
    rotate_about_axis

    Examples
    --------
    Rotation about the *z*-axis by 90 degrees:

    >>> position = np.array([2.0, 0.0, 0.0])
    >>> rotated = rotate(position, angles=(0, 0, 0.5*np.pi))
    >>> print(rotated)
    [1.225e-16 2.000e+00 0.000e+00]
    """
    from numpy import sin, cos

    # Check if "angles" contains 3 angles for all dimensions
    if len(angles) != 3:
        raise ValueError("Translation vector must be container of length 3")
    # Create rotation matrices for all 3 dimensions
    rot_x = np.array([[ 1,               0,               0               ],
                      [ 0,               cos(angles[0]),  -sin(angles[0]) ],
                      [ 0,               sin(angles[0]),  cos(angles[0])  ]])
    
    rot_y = np.array([[ cos(angles[1]),  0,               sin(angles[1])  ],
                      [ 0,               1,               0               ],
                      [ -sin(angles[1]), 0,               cos(angles[1])  ]])
    
    rot_z = np.array([[ cos(angles[2]),  -sin(angles[2]), 0               ],
                      [ sin(angles[2]),  cos(angles[2]),  0               ],
                      [ 0,               0,               1               ]])
    
    positions = coord(atoms).copy()
    positions = matrix_rotate(positions, rot_z @ rot_y @ rot_x)
    
    return _put_back(atoms, positions)


def rotate_centered(atoms, angles):
    """
    Rotate the given atoms or coordinates about the *x*, *y* and *z*
    axes by given angles.
    
    The rotations are centered at the centroid of the corresponding
    structure and are performed sequentially in the order *x*, *y*, *z*.
    
    Parameters
    ----------
    atoms : Atom or AtomArray or AtomArrayStack or ndarray, shape=(3,) or shape=(n,3) or shape=(m,n,3)
        The atoms of which the coordinates are transformed.
        The coordinates can be directly provided as :class:`ndarray`.
    angles: array-like, length=3
        The rotation angles in radians around axes *x*, *y* and *z*.
    
    Returns
    -------
    transformed : Atom or AtomArray or AtomArrayStack or ndarray, shape=(3,) or shape=(n,3) or shape=(m,n,3)
        A copy of the input atoms or coordinates, rotated by the given
        angles.
        
    See Also
    --------
    rotate
    rotate_about_axis
    """
    if len(coord(atoms).shape) == 1:
        # Single value -> centered rotation does not change coordinates
        return atoms.copy()
    
    # Rotation around centroid requires moving centroid to origin
    center = coord(centroid(atoms))
    # 'centroid()' removes the second last dimesion
    # -> add dimension again to ensure correct translation
    center = center[..., np.newaxis, :]
    translated = translate(atoms, -center)
    rotated = rotate(translated, angles)
    translated_back = translate(rotated, center)
    return translated_back


def rotate_about_axis(atoms, axis, angle, support=None):
    """
    Rotate the given atoms or coordinates about a given axis by a given
    angle.
    
    Parameters
    ----------
    atoms : Atom or AtomArray or AtomArrayStack or ndarray, shape=(3,) or shape=(n,3) or shape=(m,n,3)
        The atoms of which the coordinates are transformed.
        The coordinates can be directly provided as :class:`ndarray`.
    axis : array-like, length=3
        A vector representing the direction of the rotation axis.
        The length of the vector is irrelevant.
    angle : float
        The rotation angle in radians.
    support : array-like, length=3, optional
        An optional support vector for the rotation axis, i.e. the
        center of the rotation.
        By default, the center of the rotation is at *(0,0,0)*.
    
    Returns
    -------
    transformed : Atom or AtomArray or AtomArrayStack or ndarray, shape=(3,) or shape=(n,3) or shape=(m,n,3)
        A copy of the input atoms or coordinates, rotated about the
        given axis.
        
    See Also
    --------
    rotate
    rotate_centered

    Examples
    --------
    Rotation about a custom axis on the *y*-*z*-plane by 90 degrees:

    >>> position = np.array([2.0, 0.0, 0.0])
    >>> axis = [0.0, 1.0, 1.0]
    >>> rotated = rotate_about_axis(position, axis, angle=0.5*np.pi)
    >>> print(rotated)
    [ 1.225e-16  1.414e+00 -1.414e+00]
    """
    positions = coord(atoms).copy()
    if support is not None:
        # Transform coordinates
        # so that the axis support vector is at (0,0,0)
        positions -= np.asarray(support)
    
    # Normalize axis
    axis = np.asarray(axis, dtype=np.float32).copy()
    if np.linalg.norm(axis) == 0:
        raise ValueError("Length of the rotation axis is 0")
    norm_vector(axis)
    # Save some interim values, that are used repeatedly in the
    # rotation matrix, for clarity and to save computation time
    sin_a = np.sin(angle)
    cos_a = np.cos(angle)
    icos_a = 1 - cos_a
    x = axis[...,0]
    y = axis[...,1]
    z = axis[...,2]
    # Rotation matrix is taken from
    # https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
    rot_matrix = np.array([
        [ cos_a + icos_a*x**2,  icos_a*x*y - z*sin_a,  icos_a*x*z + y*sin_a],
        [icos_a*x*y + z*sin_a,   cos_a + icos_a*y**2,  icos_a*y*z - x*sin_a],
        [icos_a*x*z - y*sin_a,  icos_a*y*z + x*sin_a,   cos_a + icos_a*z**2]
    ])

    # For proper rotation reshape into a maximum of 2 dimensions
    orig_ndim = positions.ndim
    if orig_ndim > 2:
        orig_shape = positions.shape
        positions = positions.reshape(-1, 3)
    # Apply rotation
    positions = np.dot(rot_matrix, positions.T).T
    # Reshape back into original shape
    if orig_ndim > 2:
        positions = positions.reshape(*orig_shape)

    if support is not None:
        # Transform coordinates back to original support vector position
        positions += np.asarray(support)
    
    return _put_back(atoms, positions)


def orient_principal_components(atoms, order=None):
    """
    Translate and rotate the atoms to be centered at the origin with
    the principal axes aligned to the Cartesian axes, as specified by
    the `order` parameter. By default, x, y, z.

    By default, the resulting coordinates have the highest variance in
    the x-axis and the lowest variance on the z-axis. Setting the `order`
    parameter will change the direction of the highest variance.
    For example, ``order=(2, 1, 0)`` results in highest variance along the
    z-axis and lowest along the x-axis.

    Parameters
    ----------
    atoms : AtomArray or ndarray, shape=(n,3)
        The atoms of which the coordinates are transformed.
        The coordinates can be directly provided as :class:`ndarray`.
    order : array-like, length=3
        The order of decreasing variance. Setting `order` to ``(2, 0, 1)``
        results in highest variance along the y-axis and lowest along
        the x-axis. Default = ``(0, 1, 2)``.

    Returns
    -------
    AtomArray or ndarray, shape=(n,3)
        The atoms with coordinates centered at the orgin and aligned with
        xyz axes.

    See Also
    --------
    rotate_centered
    rotate_about_axis

    Examples
    --------
    Align principal components to xyz axes (default), or specify the order
    of variance.

    >>> print("original variance =", atom_array.coord.var(axis=0))
    original variance = [26.517 20.009  9.325]
    >>> moved = orient_principal_components(atom_array)
    >>> print("moved variance =", moved.coord.var(axis=0))
    moved variance = [28.906 18.495  8.450]
    >>> # Note the increase in variance along the x-axis
    >>> # Specifying the order keyword changes the orientation
    >>> moved_z = orient_principal_components(atom_array, order=(2, 1, 0))
    >>> print("moved (zyx) variance =", moved_z.coord.var(axis=0))
    moved (zyx) variance = [ 8.450 18.495 28.906]
    """
    # Check user input
    coords = coord(atoms)
    if coords.ndim != 2:
        raise ValueError(f"Expected input shape of (n, 3), got {coords.shape}.")
    row, col = coords.shape
    if (row < 3) or (col != 3):
        raise ValueError(
            f"Expected at least 3 entries, {row} given,"
            f" and 3 dimensions, {col} given."
        )
    if order is None:
        order = np.array([0, 1, 2])
    else:
        order = np.asarray(order, dtype=int)
        if order.shape != (3,):
            raise ValueError(
                f"Expected order to have shape (3,), not {order.shape}"
            )
        if not (np.sort(order) == np.arange(3)).all():
            raise ValueError("Expected order to contain [0, 1, 2].")

    # place centroid of the atoms at the origin
    centered = coords - coords.mean(axis=0)

    # iterate a few times to ensure the ideal rotation has been applied
    identity = np.eye(3)
    MAX_ITER = 50
    for _ in range(MAX_ITER):
        # PCA, W is the component matrix, s ~ explained variance
        _, sigma, components = np.linalg.svd(centered)

        # sort 1st, 2nd, 3rd pca compontents along x, y, z
        idx = sigma.argsort()[::-1][order]
        ident = np.eye(3)[:, idx]

        # Run the Kabsch algorithm to orient principal components
        # along the x, y, z axes
        v, _, wt = np.linalg.svd(components.T @ ident)

        if np.linalg.det(v) * np.linalg.det(wt) < -0:
            v[:, -1] *= -1
        # Calculate rotation matrix
        rotation = v @ wt
        if np.isclose(rotation, identity, atol=1e-5).all():
            break
        # Apply rotation, keep molecule centered on the origin
        centered = centered @ rotation
    return _put_back(atoms, centered)


def align_vectors(atoms, origin_direction, target_direction,
                  origin_position=None, target_position=None):
    """
    Apply a transformation to atoms or coordinates, that would transfer
    a origin vector to a target vector.

    At first the transformation (translation and rotation), that is
    necessary to align the origin vector to the target vector is
    calculated.
    This means, that the application of the transformation on the
    origin vector would give the target vector.
    Then the same transformation is applied to the given
    atoms/coordinates. 
    
    Parameters
    ----------
    atoms : Atom or AtomArray or AtomArrayStack or ndarray, shape=(3,) or shape=(n,3) or shape=(m,n,3)
        The atoms of which the coordinates are transformed.
        The coordinates can be directly provided as :class:`ndarray`.
    origin_direction, target_direction : array-like, length=3
        The vectors representing the direction of the origin or target,
        respectively.
        The length of the vectors is irrelevant.
    origin_position, target_position : array-like, length=3, optional
        Optional support vectors for the origin or target, respectively.
        By default, origin and target start at *(0,0,0)*.
    
    Returns
    -------
    transformed : Atom or AtomArray or AtomArrayStack or ndarray, shape=(3,) or shape=(n,3) or shape=(m,n,3)
        A copy of the input atoms or coordinates with the applied
        transformation.
        
    See Also
    --------
    rotate
    rotate_centered

    Examples
    --------
    Align two different residues at their CA-CB bond, i.e. the CA and CB
    position of both residues should be equal:

    >>> first_residue = atom_array[atom_array.res_id == 1]
    >>> second_residue = atom_array[atom_array.res_id == 2]
    >>> first_residue_ca = first_residue[first_residue.atom_name == "CA"]
    >>> first_residue_cb = first_residue[first_residue.atom_name == "CB"]
    >>> second_residue_ca = second_residue[second_residue.atom_name == "CA"]
    >>> second_residue_cb = second_residue[second_residue.atom_name == "CB"]
    >>> first_residue = align_vectors(
    ...     first_residue,
    ...     origin_direction = first_residue_cb.coord - first_residue_ca.coord,
    ...     target_direction = second_residue_cb.coord - second_residue_ca.coord,
    ...     origin_position = first_residue_ca.coord,
    ...     target_position = second_residue_ca.coord
    ... )
    >>> # The CA and CB coordinates of both residues are now almost equal
    >>> print(first_residue)
        A       1  ASN N      N        -5.549    3.788   -1.125
        A       1  ASN CA     C        -4.923    4.002   -2.452
        A       1  ASN C      C        -3.877    2.949   -2.808
        A       1  ASN O      O        -4.051    2.292   -3.825
        A       1  ASN CB     C        -4.413    5.445   -2.618
        A       1  ASN CG     C        -5.593    6.413   -2.670
        A       1  ASN OD1    O        -6.738    5.988   -2.651
        A       1  ASN ND2    N        -5.362    7.711   -2.695
        A       1  ASN H1     H        -5.854    2.830   -1.023
        A       1  ASN H2     H        -4.902    4.017   -0.382
        A       1  ASN H3     H        -6.357    4.397   -1.047
        A       1  ASN HA     H        -5.711    3.870   -3.198
        A       1  ASN HB2    H        -3.781    5.697   -1.788
        A       1  ASN HB3    H        -3.860    5.526   -3.556
        A       1  ASN HD21   H        -4.440    8.115   -2.680
        A       1  ASN HD22   H        -6.190    8.284   -2.750
    >>> print(second_residue)
        A       2  LEU N      N        -6.379    4.031   -2.228
        A       2  LEU CA     C        -4.923    4.002   -2.452
        A       2  LEU C      C        -4.136    3.187   -1.404
        A       2  LEU O      O        -3.391    2.274   -1.760
        A       2  LEU CB     C        -4.411    5.450   -2.619
        A       2  LEU CG     C        -4.795    6.450   -1.495
        A       2  LEU CD1    C        -3.612    6.803   -0.599
        A       2  LEU CD2    C        -5.351    7.748   -2.084
        A       2  LEU H      H        -6.821    4.923   -2.394
        A       2  LEU HA     H        -4.750    3.494   -3.403
        A       2  LEU HB2    H        -3.340    5.414   -2.672
        A       2  LEU HB3    H        -4.813    5.817   -3.564
        A       2  LEU HG     H        -5.568    6.022   -0.858
        A       2  LEU HD11   H        -3.207    5.905   -0.146
        A       2  LEU HD12   H        -2.841    7.304   -1.183
        A       2  LEU HD13   H        -3.929    7.477    0.197
        A       2  LEU HD21   H        -4.607    8.209   -2.736
        A       2  LEU HD22   H        -6.255    7.544   -2.657
        A       2  LEU HD23   H        -5.592    8.445   -1.281
    """
    origin_direction = np.asarray(
        origin_direction, dtype=np.float32
    ).squeeze()
    target_direction = np.asarray(
        target_direction, dtype=np.float32
    ).squeeze()
    # check that original and target direction are vectors of shape (3,)
    if origin_direction.shape != (3,):
        raise ValueError(
            f"Expected origin vector to have shape (3,), "
            f"got {origin_direction.shape}"
        )
    if target_direction.shape != (3,):
        raise ValueError(
            f"Expected target vector to have shape (3,), "
            f"got {target_direction.shape}"
        )
    if np.linalg.norm(origin_direction) == 0:
        raise ValueError("Length of the origin vector is 0")
    if np.linalg.norm(target_direction) == 0:
        raise ValueError("Length of the target vector is 0")
    if origin_position is not None: 
        origin_position = np.asarray(origin_position, dtype=np.float32)
    if target_position is not None: 
        target_position = np.asarray(target_position, dtype=np.float32)

    positions = coord(atoms).copy()
    if origin_position is not None:
        # Transform coordinates
        # so that the position of the origin vector is at (0,0,0)
        positions -= origin_position
    
    # Normalize direction vectors
    origin_direction = origin_direction.copy()
    norm_vector(origin_direction)
    target_direction = target_direction.copy()
    norm_vector(target_direction)
    # Formula is taken from
    # https://math.stackexchange.com/questions/180418/calculate-rotation-matrix-to-align-vector-a-to-vector-b-in-3d/476311#476311
    vx, vy, vz = np.cross(origin_direction, target_direction)
    v_c = np.array([
        [  0, -vz,  vy],
        [ vz,   0, -vx],
        [-vy,  vx,   0]
    ], dtype=float)
    cos_a = vector_dot(origin_direction, target_direction)
    if np.all(cos_a == -1):
        raise ValueError(
            "Direction vectors are point into opposite directions, "
            "cannot calculate rotation matrix"
        )
    rot_matrix = np.identity(3) + v_c + (v_c @ v_c) / (1 + cos_a)
    
    positions = matrix_rotate(positions, rot_matrix)
    
    if target_position is not None:
        # Transform coordinates to position of the target vector
        positions += target_position

    return _put_back(atoms, positions)


def _put_back(input_atoms, transformed):
    """
    Put the altered coordinates back into the :class:`Atom`,
    :class:`AtomArray` or :class:`AtomArrayStack`, if the input was one
    of these types.
    """
    if isinstance(input_atoms, (Atom, AtomArray, AtomArrayStack)):
        moved_atoms = input_atoms.copy()
        moved_atoms.coord = transformed
        return moved_atoms
    else:
        return transformed


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
Functions for parsing and writing an :class:`AtomArray` from/to
*MDL* connection tables (Ctab).
"""

__name__ = "biotite.structure.io"
__author__ = "Patrick Kunzmann"
__all__ = ["read_structure_from_ctab", "write_structure_to_ctab"]

import warnings
import numpy as np
from biotite.structure.error import BadStructureError
from ..atoms import AtomArray, AtomArrayStack
from ..bonds import BondList, BondType

BOND_TYPE_MAPPING = {
    1: BondType.SINGLE,
    2: BondType.DOUBLE,
    3: BondType.TRIPLE,
    6: BondType.SINGLE,
    7: BondType.DOUBLE,
    8: BondType.ANY,
}
BOND_TYPE_MAPPING_REV = {
    BondType.SINGLE: 1,
    BondType.DOUBLE: 2,
    BondType.TRIPLE: 3,
    BondType.AROMATIC_SINGLE: 1,
    BondType.AROMATIC_DOUBLE: 2,
    BondType.ANY: 8,
}

CHARGE_MAPPING = {0: 0, 1: 3, 2: 2, 3: 1, 5: -1, 6: -2, 7: -3}
CHARGE_MAPPING_REV = {val: key for key, val in CHARGE_MAPPING.items()}


def read_structure_from_ctab(ctab_lines):
    """
    Parse a *MDL* connection table (Ctab) to obtain an
    :class:`AtomArray`. :footcite:`Dalby1992`

    Parameters
    ----------
    ctab_lines : lines of str
        The lines containing the *ctab*.
        Must begin with the *counts* line and end with the `M END` line

    Returns
    -------
    atoms : AtomArray
        This :class:`AtomArray` contains the optional ``charge``
        annotation and has an associated :class:`BondList`.

    References
    ----------

    .. footbibliography::
    """
    n_atoms, n_bonds = _get_counts(ctab_lines[0])
    atom_lines = ctab_lines[1 : 1 + n_atoms]
    bond_lines = ctab_lines[1 + n_atoms : 1 + n_atoms + n_bonds]

    atoms = AtomArray(n_atoms)
    atoms.add_annotation("charge", int)
    for i, line in enumerate(atom_lines):
        atoms.coord[i, 0] = float(line[0:10])
        atoms.coord[i, 1] = float(line[10:20])
        atoms.coord[i, 2] = float(line[20:30])
        atoms.element[i] = line[31:34].strip().upper()
        charge = CHARGE_MAPPING.get(int(line[36:39]))
        if charge is None:
            warnings.warn(
                f"Cannot handle MDL charge type {int(line[36 : 39])}, "
                f"0 is used instead"
            )
            charge = 0
        atoms.charge[i] = charge

    bond_array = np.zeros((n_bonds, 3), dtype=np.uint32)
    for i, line in enumerate(bond_lines):
        bond_type = BOND_TYPE_MAPPING.get(int(line[6:9]))
        if bond_type is None:
            warnings.warn(
                f"Cannot handle MDL bond type {int(line[6 : 9])}, "
                f"BondType.ANY is used instead"
            )
            bond_type = BondType.ANY
        bond_array[i, 0] = int(line[0:3]) - 1
        bond_array[i, 1] = int(line[3:6]) - 1
        bond_array[i, 2] = bond_type
    atoms.bonds = BondList(n_atoms, bond_array)

    return atoms


def write_structure_to_ctab(atoms, default_bond_type=8):
    """
    Convert an :class:`AtomArray` into a
    *MDL* connection table (Ctab). :footcite:`Dalby1992`

    Parameters
    ----------
    atoms : AtomArray
        The array must have an associated :class:`BondList`.

    Returns
    -------
    ctab_lines : lines of str
        The lines containing the *ctab*.
        The lines begin with the *counts* line and end with the `M END`
        .line
    default_bond_type : int
        Bond type fallback in the *Bond block* if a bond has no bond_type
        defined in *atoms* array.

    References
    ----------

    .. footbibliography::
    """
    if isinstance(atoms, AtomArrayStack):
        raise TypeError(
            "An 'AtomArrayStack' was given, "
            "but only a single model can be written"
        )
    if atoms.bonds is None:
        raise BadStructureError("Input AtomArray has no associated BondList")

    try:
        charge = atoms.charge
    except AttributeError:
        charge = np.zeros(atoms.array_length(), dtype=int)

    atom_lines = [
        f"{atoms.coord[i,0]:>10.5f}"
        f"{atoms.coord[i,1]:>10.5f}"
        f"{atoms.coord[i,2]:>10.5f}"
        f" {atoms.element[i]:>3}"
        f"  {CHARGE_MAPPING_REV.get(charge[i], 0):>3d}" + f"{0:>3d}" * 10
        for i in range(atoms.array_length())
    ]

    bond_lines = [
        f"{i+1:>3d}{j+1:>3d}"
        f"{BOND_TYPE_MAPPING_REV.get(bond_type, default_bond_type):>3d}"
        + f"{0:>3d}" * 4
        for i, j, bond_type in atoms.bonds.as_array()
    ]

    counts_line = (
        f"{len(atom_lines):>3d}{len(bond_lines):>3d}     0  0  0  0  0  1 "
        "V2000"
    )

    return [counts_line] + atom_lines + bond_lines + ["M  END"]


def _get_counts(counts_line):
    elements = counts_line.split()
    return int(elements[0]), int(elements[1])



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from os.path import join
from tempfile import gettempdir
import pytest
from biotite.application.sra import FastqDumpApp
from biotite.sequence.io.fastq import FastqFile


@pytest.mark.parametrize("custom_prefix", [False, True])
def test_fastq_dump(custom_prefix):
    prefix = join(gettempdir(), "test_fastq") if custom_prefix else None
    app = FastqDumpApp("SRR6986517", output_path_prefix=prefix)
    app.start()
    app.join()

    for sequences_and_scores in app.get_sequences():
        assert isinstance(sequences_and_scores, dict)
    
    for fastq_file in app.get_fastq():
        assert isinstance(fastq_file, FastqFile)



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import numpy as np
import pytest
import biotite.sequence as seq
from biotite.application.viennarna import RNAfoldApp
from ..util import is_not_installed


@pytest.fixture
def sample_app():
    """
    Provide a `RNAfoldApp` object, where *RNAfold* has been executed for
    a sample sequence.
    """
    sequence = seq.NucleotideSequence("CGACGTAGATGCTAGCTGACTCGATGC")
    app = RNAfoldApp(sequence)
    app.start()
    app.join()
    return app


@pytest.mark.skipif(
    is_not_installed("RNAfold"), reason="RNAfold is not installed"
)
def test_get_dot_bracket(sample_app):
    assert sample_app.get_dot_bracket() ==  "(((.((((.......)).)))))...."


@pytest.mark.skipif(
    is_not_installed("RNAfold"), reason="RNAfold is not installed"
)
def test_get_mfe(sample_app):
    assert sample_app.get_mfe() == -1.3

@pytest.mark.skipif(
    is_not_installed("RNAfold"), reason="RNAfold is not installed"
)
def test_get_base_pairs(sample_app):
    expected_basepairs = np.array([[ 0, 22],
                                   [ 1, 21],
                                   [ 2, 20],
                                   [ 4, 19],
                                   [ 5, 18],
                                   [ 6, 16],
                                   [ 7, 15]])
    assert np.all(sample_app.get_base_pairs() == expected_basepairs)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.database.uniprot"
__author__ = "Maximilian Greil"
__all__ = ["fetch"]

from os.path import isdir, isfile, join, getsize
import os
import io
import requests
from .check import assert_valid_response

_fetch_url = "https://www.uniprot.org/"


def _get_database_name(id):
    """
    Get the correct UniProt database from the ID of the file to be downloaded.

    Parameters
    ----------
    id: str
        ID of the file to be downloaded.

    Returns
    -------
    name : str
        E-utility UniProt database name.
    """
    if id[:3] == "UPI":
        return "uniparc"
    elif id[:6] == "UniRef":
        return "uniref"
    return "uniprot"


def fetch(ids, format, target_path=None,
          overwrite=False, verbose=False):
    """
    Download files from the UniProt in various formats.

    Available databases are UniProtKB, UniRef and UniParc.

    This function requires an internet connection.

    Parameters
    ----------
    ids : str or iterable object of str
        A single ID or a list of IDs of the file(s)
        to be downloaded.
    format : {'fasta', 'gff', 'txt', 'xml', 'rdf', 'tab'}
        The format of the files to be downloaded.
    target_path : str, optional
        The target directory of the downloaded files.
        By default, the file content is stored in a file-like object
        (`StringIO` or `BytesIO`, respectively).
    overwrite : bool, optional
        If true, existing files will be overwritten. Otherwise the
        respective file will only be downloaded if the file does not
        exist yet in the specified target directory or if the file is
        empty. (Default: False)
    verbose: bool, optional
        If true, the function will output the download progress.
        (Default: False)

    Returns
    -------
    files : str or StringIO or BytesIO or list of (str or StringIO or BytesIO)
        The file path(s) to the downloaded files.
        If a single string (a single ID) was given in `ids`,
        a single string is returned. If a list (or other iterable
        object) was given, a list of strings is returned.
        If no `target_path` was given, the file contents are stored in
        either `StringIO` or `BytesIO` objects.

    Examples
    --------

    >>> import os.path
    >>> file = fetch("P12345", "fasta", path_to_directory)
    >>> print(os.path.basename(file))
    P12345.fasta
    >>> files = fetch(["P12345", "Q8K9I1"], "fasta", path_to_directory)
    >>> print([os.path.basename(file) for file in files])
    ['P12345.fasta', 'Q8K9I1.fasta']
    """
    # If only a single ID is present,
    # put it into a single element list
    if isinstance(ids, str):
        ids = [ids]
        single_element = True
    else:
        single_element = False
    # Create the target folder, if not existing
    if target_path is not None and not isdir(target_path):
        os.makedirs(target_path)
    files = []
    for i, id in enumerate(ids):
        db_name = _get_database_name(id)
        # Verbose output
        if verbose:
            print(f"Fetching file {i + 1:d} / {len(ids):d} ({id})...",
                  end="\r")
        # Fetch file from database
        if target_path is not None:
            file = join(target_path, id + "." + format)
        else:
            # 'file = None' -> store content in a file-like object
            file = None
        if file is None \
                or not isfile(file) \
                or getsize(file) == 0 \
                or overwrite:
            if format in ["fasta", "gff", "txt", "xml", "rdf", "tab"]:
                r = requests.get(_fetch_url + db_name + "/" + id + "." + format)
                content = r.text
                assert_valid_response(r.status_code)
            else:
                raise ValueError(f"Format '{format}' is not supported")
            if file is None:
                file = io.StringIO(content)
            else:
                with open(file, "w+") as f:
                    f.write(content)
        files.append(file)
    if verbose:
        print("\nDone")
    # If input was a single ID, return only a single path
    if single_element:
        return files[0]
    else:
        return files



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import pytest
import numpy as np
import biotite.structure as struc
import biotite.structure.io as strucio
from os.path import join
from ..util import data_dir


@pytest.fixture
def nuc_sample_array():
    """
    Sample structure.
    """
    nuc_sample_array = strucio.load_structure(
        join(data_dir("structure"), "4p5j.cif")
    )
    return nuc_sample_array[struc.filter_nucleotides(nuc_sample_array)]

@pytest.fixture
def expected_output():
    """
    Output for sample structure
    """
    return [
        ".[(((((.[{...)))))(((((((.......)))))))...(((((]}.)..))))[[[...(((((("
        "]]].].))))))(.)",
        ".[(((((.{[...)))))(((((((.......)))))))...(((((}].)..))))[[[...(((((("
        "]]].].))))))(.)"
    ]

@pytest.fixture
def basepair_residue_positions():
    """
    The base pairs in the sample array by their residue postions.
    """
    return np.array(
        [[1, 73],
         [2, 17],
         [3, 16],
         [4, 15],
         [5, 14],
         [6, 13],
         [8, 47],
         [9, 48],
         [18, 38],
         [19, 37],
         [20, 36],
         [21, 35],
         [22, 34],
         [23, 33],
         [24, 32],
         [42, 56],
         [43, 55],
         [44, 54],
         [45, 53],
         [46, 50],
         [57, 71],
         [58, 70],
         [59, 69],
         [63, 80],
         [64, 79],
         [65, 78],
         [66, 77],
         [67, 76],
         [68, 75],
         [81, 83]]
    )

def verify_dot_bracket_notation(output, expected_output):
    """
    Ensure that the dot_bracket notation matches a reference.
    """
    # Check that each solution is a correct output
    for solution in output:
        assert solution in expected_output

    # Check that each solution is unique
    unique_solutions = set(output)
    assert len(output) == len(unique_solutions)

def test_dot_bracket_from_structure(nuc_sample_array, expected_output):
    """
    Check the output of ``dot_bracket_from_structure()``.
    """
    output = struc.dot_bracket_from_structure(nuc_sample_array)
    verify_dot_bracket_notation(output, expected_output)

def test_dot_bracket(basepair_residue_positions, expected_output):
    """
    Check the output of ``dot_bracket()``.
    """
    output = struc.dot_bracket(
        basepair_residue_positions, len(expected_output[0])
    )
    verify_dot_bracket_notation(output, expected_output)

def test_base_pairs_from_dot_bracket(
    basepair_residue_positions, expected_output
):
    """
    Ensure that the base pairs are correctly extracted from the
    DBL-notation
    """
    for notation in expected_output:
        test_residue_positions = struc.base_pairs_from_dot_bracket(notation)
        assert np.all(test_residue_positions == basepair_residue_positions)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import glob
import itertools
from os.path import join
import numpy as np
import pytest
from pytest import approx
import biotite
import biotite.sequence as seq
import biotite.structure as struc
import biotite.structure.io.pdbx as pdbx
from ..util import data_dir


def test_get_model_count():
    pdbx_file = pdbx.PDBxFile.read(join(data_dir("structure"), "1l2y.cif"))
    test_model_count = pdbx.get_model_count(pdbx_file)
    ref_model_count = pdbx.get_structure(pdbx_file).stack_depth()
    assert test_model_count == ref_model_count


@pytest.mark.parametrize(
    "category, key, exp_value",
    [
        (
            "audit_author",
            "name",
            ["Neidigh, J.W.", "Fesinmeyer, R.M.", "Andersen, N.H."],
        ),
        ("struct_ref_seq", "pdbx_PDB_id_code", "1L2Y"),
        (
            "pdbx_nmr_ensemble",
            "conformer_selection_criteria",
            "structures with acceptable covalent geometry, "
            "structures with the least restraint violations",
        ),
    ],
)
def test_parsing(category, key, exp_value):
    pdbx_file = pdbx.PDBxFile.read(join(data_dir("structure"), "1l2y.cif"))
    cat_dict = pdbx_file[category]
    value = cat_dict[key]
    if isinstance(value, np.ndarray):
        assert value.tolist() == exp_value
    else:
        assert value == exp_value


@pytest.mark.parametrize(
    "string, use_array",
    itertools.product(["", " ", "\n", "\t"], [False, True]),
)
def test_empty_values(string, use_array):
    """
    Test whether empty strings for field values are properly replaced
    by ``'.'``.
    """
    LENGTH = 10
    ref_value = np.full(LENGTH, string, dtype="U1") if use_array else ""
    pdbx_file = pdbx.PDBxFile()
    pdbx_file.set_category(
        category="test_category",
        block="test",
        category_dict={"test_field": ref_value},
    )

    test_value = pdbx_file["test_category"]["test_field"]

    if use_array:
        assert test_value.tolist() == ["."] * LENGTH
    else:
        assert test_value == "."


@pytest.mark.parametrize(
    "path, model",
    itertools.product(
        glob.glob(join(data_dir("structure"), "*.cif")), [None, 1, -1]
    ),
)
def test_conversion(path, model):
    pdbx_file = pdbx.PDBxFile.read(path)

    try:
        array1 = pdbx.get_structure(pdbx_file, model=model)
    except biotite.InvalidFileError:
        if model is None:
            # The file cannot be parsed into an AtomArrayStack,
            # as the models contain different numbers of atoms
            # -> skip this test case
            return
        else:
            raise

    pdbx_file = pdbx.PDBxFile()
    pdbx.set_structure(pdbx_file, array1, data_block="test")

    # Remove one optional auth section in label to test fallback to label
    # fields
    atom_cat = pdbx_file.get_category("atom_site", "test")
    atom_cat.pop("auth_atom_id")
    pdbx_file.set_category("atom_site", atom_cat, "test")

    array2 = pdbx.get_structure(pdbx_file, model=model)

    assert array1.array_length() > 0
    if array1.box is not None:
        assert np.allclose(array1.box, array2.box)
    assert array1.bonds == array2.bonds
    for category in array1.get_annotation_categories():
        assert (
            array1.get_annotation(category).tolist()
            == array2.get_annotation(category).tolist()
        )
    assert array1.coord.tolist() == array2.coord.tolist()


def test_extra_fields():
    path = join(data_dir("structure"), "1l2y.cif")
    pdbx_file = pdbx.PDBxFile.read(path)
    stack1 = pdbx.get_structure(
        pdbx_file, extra_fields=["atom_id", "b_factor", "occupancy", "charge"]
    )
    pdbx_file = pdbx.PDBxFile()
    pdbx.set_structure(pdbx_file, stack1, data_block="test")
    stack2 = pdbx.get_structure(
        pdbx_file, extra_fields=["atom_id", "b_factor", "occupancy", "charge"]
    )
    assert stack1 == stack2

    path = join(data_dir("structure"), "1l2y.cif")
    pdbx_file = pdbx.PDBxFile.read(path)
    stack1 = pdbx.get_structure(
        pdbx_file, extra_fields=["atom_id", "b_factor", "occupancy", "charge"]
    )

    pdbx_file = pdbx.PDBxFile()
    pdbx.set_structure(pdbx_file, stack1, data_block="test")

    stack2 = pdbx.get_structure(
        pdbx_file, extra_fields=["atom_id", "b_factor", "occupancy", "charge"]
    )

    assert stack1.ins_code.tolist() == stack2.ins_code.tolist()
    assert stack1.atom_id.tolist() == stack2.atom_id.tolist()
    assert stack1.b_factor.tolist() == approx(stack2.b_factor.tolist())
    assert stack1.occupancy.tolist() == approx(stack2.occupancy.tolist())
    assert stack1.charge.tolist() == stack2.charge.tolist()
    assert stack1 == stack2


def test_unequal_lengths():
    valid_category_dict = {"foo1": ["1", "2", "3"], "foo2": ["1", "2", "3"]}
    # Arrays have unequal lengths -> invalid
    invalid_category_dict = {
        "foo1": ["1", "2", "3"],
        "foo2": ["1", "2", "3", "4"],
    }
    pdbx_file = pdbx.PDBxFile()
    pdbx_file.set_category("test", valid_category_dict, block="test_block")
    with pytest.raises(ValueError):
        pdbx_file.set_category(
            "test", invalid_category_dict, block="test_block"
        )


def test_list_assemblies():
    """
    Test the :func:`list_assemblies()` function based on a known
    example.
    """
    path = join(data_dir("structure"), "1f2n.cif")
    pdbx_file = pdbx.PDBxFile.read(path)

    assembly_list = pdbx.list_assemblies(pdbx_file)
    assert assembly_list == {
        "1": "complete icosahedral assembly",
        "2": "icosahedral asymmetric unit",
        "3": "icosahedral pentamer",
        "4": "icosahedral 23 hexamer",
        "5": "icosahedral asymmetric unit, std point frame",
        "6": "crystal asymmetric unit, crystal frame",
    }


@pytest.mark.parametrize("model", [None, 1, -1])
def test_get_assembly(model):
    """
    Test whether the :func:`get_assembly()` function produces the same
    number of peptide chains as the
    ``_pdbx_struct_assembly.oligomeric_count`` field indicates.
    Furthermore, check if the number of atoms in the entire assembly
    is a multiple of the numbers of atoms in a monomer.
    """

    path = join(data_dir("structure"), "1f2n.cif")
    pdbx_file = pdbx.PDBxFile.read(path)

    assembly_category = pdbx_file.get_category(
        "pdbx_struct_assembly", expect_looped=True
    )
    # Test each available assembly
    for id, ref_oligomer_count in zip(
        assembly_category["id"], assembly_category["oligomeric_count"]
    ):
        print("Assembly ID:", id)
        try:
            assembly = pdbx.get_assembly(
                pdbx_file, assembly_id=id, model=model
            )
        except biotite.InvalidFileError:
            if model is None:
                # The file cannot be parsed into an AtomArrayStack,
                # as the models contain different numbers of atoms
                # -> skip this test case
                return
            else:
                raise
        protein_assembly = assembly[..., struc.filter_amino_acids(assembly)]
        test_oligomer_count = struc.get_chain_count(protein_assembly)

        if model is None:
            assert isinstance(assembly, struc.AtomArrayStack)
        else:
            assert isinstance(assembly, struc.AtomArray)
        assert test_oligomer_count == int(ref_oligomer_count)

        # The atom count of the entire assembly should be a multiple
        # a monomer,
        monomer_atom_count = pdbx.get_structure(pdbx_file).array_length()
        assert assembly.array_length() % monomer_atom_count == 0


def test_get_sequence():
    file = pdbx.PDBxFile.read(join(data_dir("structure"), "5ugo.cif"))
    sequences = pdbx.get_sequence(file)
    file = pdbx.PDBxFile.read(join(data_dir("structure"), "4gxy.cif"))
    sequences += pdbx.get_sequence(file)
    assert str(sequences[0]) == "CCGACGGCGCATCAGC"
    assert type(sequences[0]) is seq.NucleotideSequence
    assert str(sequences[1]) == "GCTGATGCGCC"
    assert type(sequences[1]) is seq.NucleotideSequence
    assert str(sequences[2]) == "GTCGG"
    assert type(sequences[2]) is seq.NucleotideSequence
    assert (
        str(sequences[3]) == "MSKRKAPQETLNGGITDMLTELANFEKNVSQAIHKYN"
        "AYRKAASVIAKYPHKIKSGAEAKKLPGVGTKIAEKIDEFLATGKLRKLEKIRQD"
        "DTSSSINFLTRVSGIGPSAARKFVDEGIKTLEDLRKNEDKLNHHQRIGLKYFGD"
        "FEKRIPREEMLQMQDIVLNEVKKVDSEYIATVCGSFRRGAESSGDMDVLLTHPS"
        "FTSESTKQPKLLHQVVEQLQKVHFITDTLSKGETKFMGVCQLPSKNDEKEYPHR"
        "RIDIRLIPKDQYYCGVLYFTGSDIFNKNMRAHALEKGFTINEYTIRPLGVTGVA"
        "GEPLPVDSEKDIFDYIQWKYREPKDRSE"
    )
    assert type(sequences[3]) is seq.ProteinSequence
    assert (
        str(sequences[4]) == "GGCGGCAGGTGCTCCCGACCCTGCGGTCGGGAGTTAA"
        "AAGGGAAGCCGGTGCAAGTCCGGCACGGTCCCGCCACTGTGACGGGGAGTCGCC"
        "CCTCGGGATGTGCCACTGGCCCGAAGGCCGGGAAGGCGGAGGGGCGGCGAGGAT"
        "CCGGAGTCAGGAAACCTGCCTGCCGTC"
    )
    assert type(sequences[4]) is seq.NucleotideSequence



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.info"
__author__ = "Patrick Kunzmann"
__all__ = ["bond_dataset", "bond_order", "bond_type", "bonds_in_residue"]

import warnings
import copy
from os.path import join, dirname, realpath
import msgpack
from ..bonds import BondType


_intra_bonds = None


def _init_dataset():
    """
    Load the bond dataset from MessagePack file.

    Since loading the database is computationally expensive,
    this is only done, when the bond database is actually required.
    """
    global _intra_bonds
    if _intra_bonds is not None:
        # Database is already initialized
        return

    # Bonds are taken from
    # ftp://ftp.wwpdb.org/pub/pdb/data/monomers/components.cif
    # (2019/01/27)
    _info_dir = dirname(realpath(__file__))
    with open(join(_info_dir, "intra_bonds.msgpack"), "rb") as file:
        _intra_bonds= msgpack.unpack(
            file, use_list=False, raw=False, strict_map_key=False
        )


def bond_dataset():
    """
    Get a copy of the complete bond dataset extracted from the chemical 
    components dictionary.

    This dataset does only contain intra-residue bonds.

    Returns
    -------
    bonds : dict (str -> dict ((str, str) -> int))
        The bonds as nested dictionary.
        It maps residue names (up to 3-letters, upper case) to
        inner dictionaries.
        Each of these dictionary contains the bond information for the
        given residue.
        Specifically, it uses a set of two atom names, that are bonded,
        as keys and the respective :class:`BondType`
        (represented by an integer) as values.
    """
    _init_dataset()
    return copy.copy(_intra_bonds)


def bond_order(res_name, atom_name1, atom_name2):
    """
    Get the bond order for two atoms of the same residue, based
    on the PDB chemical components dictionary.

    DEPRECATED: Please use :func:`bond_type()` instead.

    Parameters
    ----------
    res_name : str
        The up to 3-letter name of the residue
        `atom_name1` and `atom_name2` belong to.
    atom_name1, atom_name2 : str
        The names of the two atoms to get the bond order from.
    
    Returns
    -------
    order : int or None
        The order of the bond between `atom_name1` and `atom_name2`.
        If the atoms form no bond, if any of the two atoms does not
        exist in the context of the residue or if the residue is unknown
        to the chemical components dictionary, `None` is returned.
    
    Examples
    --------

    >>> print(bond_order("PHE", "CA", "CB"))
    1
    >>> print(bond_order("PHE", "CG", "CD1"))
    2
    >>> print(bond_order("PHE", "CA", "CG"))
    None
    >>> print(bond_order("PHE", "FOO", "BAR"))
    None
    """
    warnings.warn("Please use `bond_type()` instead", DeprecationWarning)

    _init_dataset()
    btype = bond_type(res_name, atom_name1, atom_name2)
    if btype is None:
        return None
    elif btype == BondType.AROMATIC_SINGLE:
        return 1
    elif btype == BondType.AROMATIC_DOUBLE:
        return 2
    else:
        return int(btype)


def bond_type(res_name, atom_name1, atom_name2):
    """
    Get the :class:`BondType` for two atoms of the same residue, based
    on the PDB chemical components dictionary.

    Parameters
    ----------
    res_name : str
        The up to 3-letter name of the residue
        `atom_name1` and `atom_name2` belong to.
    atom_name1, atom_name2 : str
        The names of the two atoms to get the bond order from.
    
    Returns
    -------
    order : BondType or None
        The :class:`BondType` of the bond between `atom_name1` and
        `atom_name2`.
        If the atoms form no bond, if any of the two atoms does not
        exist in the context of the residue or if the residue is unknown
        to the chemical components dictionary, `None` is returned.
    
    Examples
    --------

    >>> print(bond_type("PHE", "CA", "CB"))
    BondType.SINGLE
    >>> print(bond_type("PHE", "CG", "CD1"))
    BondType.AROMATIC_DOUBLE
    >>> print(bond_type("PHE", "CA", "CG"))
    None
    >>> print(bond_type("PHE", "FOO", "BAR"))
    None
    """
    _init_dataset()
    group_bonds = _intra_bonds.get(res_name.upper())
    if group_bonds is None:
        return None
    # Try both atom aroders
    bond_type_int = group_bonds.get(
        (atom_name1, atom_name2),
        group_bonds.get((atom_name2, atom_name1))
    )
    if bond_type_int is not None:
        return BondType(bond_type_int)
    else:
        return None


def bonds_in_residue(res_name):
    """
    Get a dictionary containing all atoms inside a given residue
    that form a bond.

    Parameters
    ----------
    res_name : str
        The up to 3-letter name of the residue to get the bonds for.
    
    Returns
    -------
    bonds : dict (str -> int)
        A dictionary that maps tuples of two atom names to their
        respective bond types (represented as integer).
        `None` if the residue is unknown to the
        chemical components dictionary.
    
    Examples
    --------
    >>> bonds = bonds_in_residue("PHE")
    >>> for atoms, bond_type_int in sorted(bonds.items()):
    ...     atom1, atom2 = sorted(atoms)
    ...     print(f"{atom1:3} + {atom2:3} -> {str(BondType(bond_type_int))}")
    C   + O   -> BondType.DOUBLE
    C   + OXT -> BondType.SINGLE
    C   + CA  -> BondType.SINGLE
    CA  + CB  -> BondType.SINGLE
    CA  + HA  -> BondType.SINGLE
    CB  + CG  -> BondType.SINGLE
    CB  + HB2 -> BondType.SINGLE
    CB  + HB3 -> BondType.SINGLE
    CD1 + CE1 -> BondType.AROMATIC_SINGLE
    CD1 + HD1 -> BondType.SINGLE
    CD2 + CE2 -> BondType.AROMATIC_DOUBLE
    CD2 + HD2 -> BondType.SINGLE
    CE1 + CZ  -> BondType.AROMATIC_DOUBLE
    CE1 + HE1 -> BondType.SINGLE
    CE2 + CZ  -> BondType.AROMATIC_SINGLE
    CE2 + HE2 -> BondType.SINGLE
    CD1 + CG  -> BondType.AROMATIC_DOUBLE
    CD2 + CG  -> BondType.AROMATIC_SINGLE
    CZ  + HZ  -> BondType.SINGLE
    CA  + N   -> BondType.SINGLE
    H   + N   -> BondType.SINGLE
    H2  + N   -> BondType.SINGLE
    HXT + OXT -> BondType.SINGLE
    """
    _init_dataset()
    return copy.copy(_intra_bonds.get(res_name.upper()))


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import itertools
from os.path import join
import numpy as np
import pytest
import biotite.structure as struc
import biotite.structure.io.npz as npz
from ..util import data_dir


@pytest.fixture(
    params=itertools.product(
        [1, 2, 3],    # ndim
        [False, True] # as_coord
    )
)
def input_atoms(request):
    ndim, as_coord = request.param

    file = npz.NpzFile.read(join(data_dir("structure"), "1l2y.npz"))
    atoms = file.get_structure()
    
    if ndim == 2:
        # Only one model
        atoms = atoms[0]
    elif ndim == 1:
        # Only one atom
        atoms = atoms[0,0]
    
    if as_coord:
        return atoms.coord
    else:
        return atoms


@pytest.mark.parametrize("ndim", [1, 2, 3])
@pytest.mark.parametrize("as_list", [False, True])
@pytest.mark.parametrize("random_seed", np.arange(5))
def test_translate(input_atoms, ndim, as_list, random_seed):
    """
    Translate and translate back and check if the coordinates are still
    the same.
    """
    if ndim > len(input_atoms.shape):
        # Cannot run tests if translation vector has more dimensions
        # as input coordinates/atoms
        return
    
    np.random.seed(random_seed)
    vectors = np.random.rand(*struc.coord(input_atoms).shape[-ndim:])
    vectors *= 10
    neg_vectors = -vectors
    if as_list:
        vectors = vectors.tolist()
        neg_vectors = neg_vectors.tolist()
    
    translated = struc.translate(input_atoms, vectors)
    restored = struc.translate(translated, neg_vectors)

    assert type(restored) == type(input_atoms)
    assert struc.coord(restored).shape == struc.coord(input_atoms).shape
    assert np.allclose(
        struc.coord(restored), struc.coord(input_atoms), atol=1e-5
    )


@pytest.mark.parametrize("as_list", [False, True])
@pytest.mark.parametrize("axis", [0, 1, 2]) # x, y, z
@pytest.mark.parametrize("random_seed", np.arange(5))
@pytest.mark.parametrize("centered", [False, True])
def test_rotate(input_atoms, as_list, axis, random_seed, centered):
    """
    Rotate and rotate back and check if the coordinates are still
    the same.
    """
    np.random.seed(random_seed)
    angles = np.zeros(3)
    angles[axis] = np.random.rand() * 2*np.pi
    neg_angles = -angles
    if as_list:
        angles = angles.tolist()
        neg_angles = neg_angles.tolist()
    
    func = struc.rotate_centered if centered else struc.rotate
    rotated = func(input_atoms, angles)
    restored = func(rotated, neg_angles)

    assert type(restored) == type(input_atoms)
    assert struc.coord(restored).shape == struc.coord(input_atoms).shape
    print(np.max(np.abs(struc.coord(restored) - struc.coord(input_atoms))))
    assert np.allclose(
        struc.coord(restored), struc.coord(input_atoms), atol=1e-5
    )
    if centered and struc.coord(input_atoms).ndim > 1:
        assert np.allclose(
            struc.centroid(restored), struc.centroid(input_atoms), atol=1e-5
        )


@pytest.mark.parametrize("x", [0, 2*np.pi])
@pytest.mark.parametrize("y", [0, 2*np.pi])
@pytest.mark.parametrize("z", [0, 2*np.pi])
@pytest.mark.parametrize("centered", [False, True])
def test_rotate_360(input_atoms, x, y, z, centered):
    """
    Rotate by 360 degrees and expect that the coordinates have not
    changed.
    """
    func = struc.rotate_centered if centered else struc.rotate
    rotated = func(input_atoms, [x, y, z])
    
    assert type(rotated) == type(input_atoms)
    assert struc.coord(rotated).shape == struc.coord(input_atoms).shape
    assert np.allclose(
        struc.coord(rotated), struc.coord(input_atoms), atol=1e-5
    )
    if centered and struc.coord(input_atoms).ndim > 1:
        assert np.allclose(
            struc.centroid(rotated), struc.centroid(input_atoms), atol=1e-5
        )


@pytest.mark.parametrize("ndim", [1, 2, 3])
def test_rotate_known(ndim):
    """
    Rotate a vector at the Y-axis about the X-axis by 90 degrees and
    expect a rotated vector at the Z-axis.
    """
    shape = (1,) * (ndim-1) + (3,)
    vector = np.zeros(shape)
    vector[...] = [0, 1, 0]

    exp_rotated = np.zeros(shape)
    exp_rotated[...] = [0, 0, 1]

    # Rotation by 90 degrees
    test_rotated = struc.rotate(vector, [0.5 * np.pi, 0, 0])
    
    assert test_rotated.shape == exp_rotated.shape
    assert np.allclose(test_rotated, exp_rotated, atol=1e-5)


@pytest.mark.parametrize("axis", [0, 1, 2]) # x, y, z
@pytest.mark.parametrize("random_seed", np.arange(5))
def test_rotate_measure(axis, random_seed):
    """
    Rotate and measure resulting angle that should be equal to the input
    angle.
    """
    np.random.seed(random_seed)
    # Maximum rotation is only 180 degrees,
    # as with higher angles the measured angle would decrease again
    ref_angle = np.random.rand() * np.pi
    angles = np.zeros(3)
    angles[axis] = ref_angle

    # The measured angle is only equal to the input angle,
    # if the input coordinates have no component on the rotation axis 
    input_coord = np.ones(3)
    input_coord[axis] = 0

    rotated = struc.rotate(input_coord, angles)
    test_angle = struc.angle(rotated, 0, input_coord)
    
    # Vector length should be unchanged
    assert np.linalg.norm(rotated) \
        == pytest.approx(np.linalg.norm(input_coord))
    assert test_angle == pytest.approx(ref_angle)


@pytest.mark.parametrize("as_list", [False, True])
@pytest.mark.parametrize("use_support", [False, True])
@pytest.mark.parametrize("random_seed", np.arange(5))
def test_rotate_about_axis(input_atoms, as_list, use_support, random_seed):
    """
    Rotate and rotate back and check if the coordinates are still
    the same.
    """
    np.random.seed(random_seed)
    axis = np.random.rand(3)
    angle = np.random.rand()
    neg_angle = -angle
    support = np.random.rand(3) if use_support else None
    if as_list:
        axis = axis.tolist()
        support = support.tolist() if support is not None else None

    rotated = struc.rotate_about_axis(input_atoms, axis, angle, support)
    restored = struc.rotate_about_axis(rotated, axis, neg_angle, support)

    assert type(restored) == type(input_atoms)
    assert struc.coord(restored).shape == struc.coord(input_atoms).shape
    assert np.allclose(
        struc.coord(restored), struc.coord(input_atoms), atol=1e-5
    )


@pytest.mark.parametrize("axis", [0, 1, 2]) # x, y, z
@pytest.mark.parametrize("random_seed", np.arange(5))
def test_rotate_about_axis_consistency(input_atoms, axis, random_seed):
    """
    Compare the outcome of :func:`rotate_about_axis()` with
    :func:`rotate()`.
    """
    np.random.seed(random_seed)
    angle = np.random.rand() * 2 * np.pi
    
    angles = np.zeros(3)
    angles[axis] = angle
    ref_rotated = struc.rotate(input_atoms, angles)

    rot_axis = np.zeros(3)
    # Length of axis should be irrelevant
    rot_axis[axis] = np.random.rand()
    test_rotated = struc.rotate_about_axis(input_atoms, rot_axis, angle,)

    assert type(test_rotated) == type(ref_rotated)
    assert struc.coord(test_rotated).shape == struc.coord(ref_rotated).shape
    assert np.allclose(
        struc.coord(test_rotated), struc.coord(ref_rotated), atol=1e-5
    )


@pytest.mark.parametrize("random_seed", np.arange(5))
@pytest.mark.parametrize("use_support", [False, True])
def test_rotate_about_axis_360(input_atoms, random_seed, use_support):
    """
    Rotate by 360 degrees around an arbitrary axis and expect that the
    coordinates have not changed.
    """
    np.random.seed(random_seed)
    axis = np.random.rand(3)
    support  = np.random.rand(3) if use_support else None
    
    rotated = struc.rotate_about_axis(input_atoms, axis, 2*np.pi, support)

    assert type(rotated) == type(input_atoms)
    assert struc.coord(rotated).shape == struc.coord(input_atoms).shape
    assert np.allclose(
        struc.coord(rotated), struc.coord(input_atoms), atol=1e-5
    )


@pytest.mark.parametrize("as_list", [False, True])
@pytest.mark.parametrize("order", (
    np.array([0, 1, 2]),
    np.array([0, 2, 1]),
    np.array([1, 0, 2]),
    np.array([2, 0, 1]),
    np.array([2, 1, 0]),
    np.array([1, 2, 0]),
))
def test_orient_principal_components(input_atoms, as_list, order):
    """
    Orient atoms such that the variance in each axis is greatest
    where the value for `order` is smallest.

    The number of dimensions in input_atoms must be 2.
    """
    if struc.coord(input_atoms).ndim != 2:
        with pytest.raises(ValueError):
            struc.orient_principal_components(input_atoms)
        return

    if as_list:
        order = order.tolist()

    result = struc.orient_principal_components(input_atoms, order=order)
    neg_variance = -struc.coord(result).var(axis=0)
    assert type(result) == type(input_atoms)
    assert (neg_variance.argsort() == np.argsort(order)).all()


@pytest.mark.parametrize("bad_order", (np.array([2, 3, 4]), np.arange(4)))
def test_orient_principal_components_bad_order(input_atoms, bad_order):
    """
    Ensure correct inputs for `order`.
    """
    with pytest.raises(ValueError):
        struc.orient_principal_components(input_atoms, order=bad_order)


@pytest.mark.parametrize("as_list", [False, True])
@pytest.mark.parametrize("use_support", [False, True])
@pytest.mark.parametrize("random_seed", np.arange(5))
def test_align_vectors(input_atoms, as_list, use_support, random_seed):
    """
    Align, swap alignment vectors and align again.
    Expect that the original coordinates have been restored.
    """
    np.random.seed(random_seed)
    source_direction = np.random.rand(3)
    target_direction = np.random.rand(3)
    if use_support:
        source_position  = np.random.rand(3)
        target_position  = np.random.rand(3)
    else:
        source_position = None
        target_position = None
    
    if as_list:
        source_direction = source_direction.tolist()
        target_direction = target_direction.tolist()
        if use_support:
            source_position = source_position.tolist()
            target_position = target_position.tolist()
    
    transformed = struc.align_vectors(
        input_atoms,
        source_direction, target_direction,
        source_position, target_position
    )
    restored = struc.align_vectors(
        transformed,
        target_direction, source_direction, 
        target_position, source_position
    )

    assert type(restored) == type(input_atoms)
    assert struc.coord(restored).shape == struc.coord(input_atoms).shape
    assert np.allclose(
        struc.coord(restored), struc.coord(input_atoms), atol=1e-5
    )


def test_align_vectors_non_vector_inputs(input_atoms):
    """
    Ensure input vectors to ``struct.align_vectors`` have the correct shape.
    """
    source_direction = np.random.rand(2, 3)
    target_direction = np.random.rand(2, 3)
    with pytest.raises(ValueError):
        struc.align_vectors(
            input_atoms,
            source_direction,
            target_direction,
        )
    source_direction = np.random.rand(4)
    target_direction = np.random.rand(4)
    with pytest.raises(ValueError):
        struc.align_vectors(
            input_atoms,
            source_direction,
            target_direction,
        )
